import logging

from mstrio import config
from mstrio.api import facts
from mstrio.connection import Connection
from mstrio.modeling.expression import ExpressionFormat, FactExpression
from mstrio.modeling.schema.helpers import (
    DataType,
    ObjectSubType,
    SchemaObjectReference,
)
from mstrio.object_management import search_operations
from mstrio.object_management.folder import Folder
from mstrio.object_management.search_enums import SearchPattern
from mstrio.types import ObjectTypes
from mstrio.users_and_groups import User
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, MoveMixin
from mstrio.utils.enum_helper import get_enum_val
from mstrio.utils.helper import (
    Dictable,
    filter_params_for_func,
    find_object_with_name,
    get_valid_project_id,
)
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.translation_mixin import TranslationMixin
from mstrio.utils.version_helper import class_version_handler, method_version_handler

logger = logging.getLogger(__name__)


@method_version_handler('11.3.0100')
def list_facts(
    connection: Connection,
    name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    project_id: str | None = None,
    project_name: str | None = None,
    search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
    show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    **filters,
) -> list["Fact"] | list[dict]:
    """Get list of Fact objects or dicts with them.

    Specify either `project_id` or `project_name`.
    When `project_id` is provided (not `None`), `project_name` is omitted.

    Note:
        When `project_id` is `None` and `project_name` is `None`,
        then its value is overwritten by `project_id` from `connection` object.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        name (optional, str): value the search pattern is set to, which
            will be applied to the names of facts being searched
        to_dictionary (optional, bool): If `True` returns dictionaries, by
            default (`False`) returns `Fact` objects.
        limit (optional, int): limit the number of elements returned. If `None`
            (default), all objects are returned.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        search_pattern (SearchPattern enum or int, optional): pattern to
            search for, such as Begin With or Exactly. Possible values are
            available in ENUM `mstrio.object_management.SearchPattern`.
            Default value is CONTAINS (4).
        show_expression_as (optional, enum or str): specify how expressions
            should be presented.
            Available values:
                - `ExpressionFormat.TREE` or `tree` (default)
                - `ExpressionFormat.TOKENS or `tokens`
        **filters: Available filter parameters: ['id', 'name',
            'type', 'subtype', 'date_created', 'date_modified', 'version',
            'acg', 'owner', 'ext_type']

    Returns:
        list of fact objects or list of fact dictionaries.
    """
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=not project_name,
    )

    objects = search_operations.full_search(
        connection,
        object_types=ObjectTypes.FACT,
        project=project_id,
        name=name,
        pattern=search_pattern,
        limit=limit,
        **filters,
    )
    if to_dictionary:
        return objects
    else:
        show_expression_as = (
            show_expression_as
            if isinstance(show_expression_as, ExpressionFormat)
            else ExpressionFormat(show_expression_as)
        )
        return [
            Fact.from_dict(
                source={'show_expression_as': show_expression_as, **obj},
                connection=connection,
                with_missing_value=True,
            )
            for obj in objects
        ]


@class_version_handler('11.3.0100')
class Fact(Entity, CopyMixin, DeleteMixin, MoveMixin, TranslationMixin):
    """Python representation for Microstrategy `Fact` object.

    Attributes:
        id: fact's ID
        name: fact's name
        sub_type: string literal used to identify the type of a metadata object
        description: fact's description
        type: object type, `ObjectTypes` enum
        subtype: object subtype, `ObjectSupTypes` enum
        ext_type: object extended type, `ExtendedType` enum
        ancestors: list of ancestor folders
        date_created: creation time, `DateTime` object
        date_modified: last modification time `DateTime` object
        destination_folder_id: a globally unique identifier used to distinguish
            between metadata objects within the same project
        is_embedded: if `True` indicates that the target object of this
            reference is embedded within this object. Alternatively if this
            object is itself embedded, then it means that the target object is
            embedded in the same container as this object
        data_type: dataype that the engine should use to store a intermediate
            values for this fact, `DataType` object
        expressions: Array with a member object for each separately defined
            expression currently in use by a fact. Often a fact expression takes
            the form of just a single column name, but more complex expressions
            are possible.
        entry_level: this is the natural level (in other words a set of
            attributes) of the fact. This property is not set by the architects.
            Instead it is deduced by the platform by examining the specification
            of the fact, and considering the attributes and the relationships
            between them. Because it can only be computed by considering the
            schema as a whole, the value is not automatically updated each time
            the fact is modified. It will be correct following a schema update
            operation.
        owner: `User` object that is the owner
        acg: access rights (see `EnumDSSXMLAccessRightFlags` for possible
            values)
        acl: object access control list
        version_id: the version number this object is currently carrying
        hidden: Specifies whether the object is hidden
    """

    _OBJECT_TYPE = ObjectTypes.FACT

    _API_GETTERS = {
        (
            'id',
            'sub_type',
            'name',
            'is_embedded',
            'description',
            'destination_folder_id',
            'path',
            'data_type',
            'expressions',
            'entry_level',
        ): facts.read_fact,
        (
            'abbreviation',
            'type',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'target_info',
            'hidden',
            'comments',
        ): objects_processors.get_info,
    }
    _API_PATCH = {
        ('data_type', 'expressions'): (facts.update_fact, 'partial_put'),
        ('name', 'description', 'folder_id', 'hidden', 'comments'): (
            objects_processors.update,
            'partial_put',
        ),
    }

    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': User.from_dict,
        'data_type': DataType.from_dict,
        'expressions': (
            lambda source, connection: [
                FactExpression.from_dict(content, connection) for content in source
            ]
        ),
        'entry_level': (
            lambda source, connection: [
                SchemaObjectReference.from_dict(content, connection)
                for content in source
            ]
        ),
    }

    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        name: str | None = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    ) -> None:
        """Initialize a new instance of Fact class.

        Note:
            Parameter `name` is not used when fetching. If only `name` parameter
            is provided, `id` will be found automatically if such object exists.

        Args:
             connection (object): MicroStrategy connection object returned by
                `connection.Connection()`.
            id (optional, str): Identifier of a pre-existing fact containing the
                required data.
            name (optional, str): name of a fact object containing
                the required data.
            show_expression_as (optional, enum or str): specify how expressions
                should be presented.
                Available values:
                - `ExpressionFormat.TREE` or `tree` (default)
                - `ExpressionFormat.TOKENS or `tokens`

        Raises:
            ValueError: if both `id` and `name` are not provided, if there
                are multiple facts with given `name`, if project is not
                selected in provided `connection` or if Fact with given `name`
                doesn't exist.
        """
        connection._validate_project_selected()
        if id is None:
            if name is None:
                raise ValueError(
                    "Please specify either 'name' or 'id' parameter in the constructor."
                )

            fact = find_object_with_name(
                connection=connection,
                cls=self.__class__,
                name=name,
                listing_function=list_facts,
                search_pattern=SearchPattern.EXACTLY,
            )
            id = fact['id']
        super().__init__(
            connection=connection,
            object_id=id,
            name=name,
            show_expression_as=show_expression_as,
        )

    def _init_variables(self, default_value, **kwargs) -> None:
        """Initialize all properties when creating `Fact` object from
        a dictionary."""
        super()._init_variables(default_value=default_value, **kwargs)
        self._sub_type = kwargs.get('sub_type', default_value)
        self._is_embedded = kwargs.get('is_embedded', default_value)
        self._destination_folder_id = kwargs.get('destination_folder_id', default_value)
        data_type = kwargs.get('data_type')
        self.data_type = (
            default_value if data_type is None else DataType.from_dict(data_type)
        )
        expressions = kwargs.get('expressions')
        self._expressions = (
            default_value
            if expressions is None
            else [FactExpression.from_dict(expression) for expression in expressions]
        )
        entry_level = kwargs.get('entry_level')
        self._entry_level = (
            default_value
            if entry_level is None
            else [SchemaObjectReference.from_dict(obj) for obj in entry_level]
        )
        self._version_id = kwargs.get('version_id')
        show_expression_as = kwargs.get('show_expression_as', 'tree')
        self._show_expression_as = (
            show_expression_as
            if isinstance(show_expression_as, ExpressionFormat)
            else ExpressionFormat(show_expression_as)
        )

    @classmethod
    def create(
        cls,
        connection: Connection,
        name: str,
        destination_folder: Folder | str,
        expressions: list[FactExpression | dict],
        data_type: DataType | dict | None = None,
        description: str | None = None,
        is_embedded: bool = False,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    ) -> "Fact":
        """Create new fact object.

        Args:
            connection (object): MicroStrategy connection object returned
                by `connection.Connection()`
            name (str): new fact's name
            sub_type (str, enum): new fact's sub_type
            destination_folder (str or object): A globally unique identifier or
                unique folder name used to distinguish between metadata objects
                within the same project.
                It is possible for two metadata objects in different
                projects to have the same Object Id.
            expressions (list): List of `FactExpression` objects or dictionaries
                representing the fact object.
            data_type (object or dict): `DataType` object or dict with
                definition of data type and precision
            description (str): new fact's description
            is_embedded (optional, bool): `True` indicates that the target
                object of this reference is embedded within this object.
                Alternatively if this object is itself embedded, then it means
                that the target object is embedded in the same container as this
                object. Default value is `False`.
            show_expression_as (optional, enum or str): specify how
                expressions should be presented
                Available values:
                - `ExpressionFormat.TREE` or `tree` (default)
                - `ExpressionFormat.TOKENS` or `tokens`

        Returns:
            `Fact` class object.
        """
        if not isinstance(expressions, list):
            expressions = [expressions]
        body = {
            "information": {
                "name": name,
                "subType": ObjectSubType.FACT.value,
                "isEmbedded": is_embedded,
                "description": description,
                "destinationFolderId": (
                    destination_folder.id
                    if isinstance(destination_folder, Folder)
                    else destination_folder
                ),
            },
            "dataType": (
                data_type.to_dict() if isinstance(data_type, DataType) else data_type
            ),
            "expressions": [
                (e.to_dict() if isinstance(e, Dictable) else e) for e in expressions
            ],
            "entryLevel": [],
        }
        response = facts.create_fact(
            connection=connection,
            body=body,
            show_expression_as=get_enum_val(show_expression_as, ExpressionFormat),
        ).json()
        if config.verbose:
            logger.info(
                f"Successfully created fact named: '{name}' with ID: '"
                f"{response['id']}'."
            )

        return cls.from_dict(
            source={**response, 'show_expression_as': show_expression_as},
            connection=connection,
        )

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        data_type: DataType | dict | None = None,
        hidden: bool | None = None,
        comments: str | None = None,
    ):
        """Alter fact properties.

        Args:
            name (optional, str): fact's name
            description (optional, str): fact's description
            data_type (optional, object or dict): fact's data type definition
            hidden (bool, optional): Specifies whether the object is hidden.
                Default value: False.
            comments (optional, str): fact's long description
        """
        properties = filter_params_for_func(self.alter, locals(), exclude=['self'])
        if data_type:
            properties['expressions'] = self.expressions
        self._alter_properties(**properties)

    def get_tables(
        self, expression: FactExpression | str | None = None
    ) -> list[SchemaObjectReference]:
        """Get list of all tables in given fact expression. If expression
        argument is not specified, list all tables for fact.

        Args:
            expression (optional, str or object): id of the fact expression or
                `FactExpression` object to get tables from.

        Returns:
            List of tables in given `FactExpression` or all tables for fact.
        """
        expressions = self.expressions
        if expression:
            expression_id = (
                expression.id if isinstance(expression, FactExpression) else expression
            )
            expressions = [expr for expr in expressions if expr.id == expression_id]
        tables_list = {tab for expr in expressions for tab in expr.tables}
        return list(tables_list)

    def add_expression(self, expression: FactExpression | dict) -> None:
        """Add expression to the fact.

        Args:
            expression (optional, object or dict): `FactExpression` object or
                dictionary having all necessary information to be added to the
                fact.
        """
        expressions = self.expressions.copy()
        expressions.append(expression)
        self._alter_properties(expressions=expressions)

    def remove_expression(self, expression: str) -> None:
        """Remove expression from the fact.

        Args:
            expression (str): id of `FactExpression` object to be removed
        """
        expressions = self.expressions.copy()
        expressions = [expr for expr in expressions if expr.id != expression]
        self._alter_properties(expressions=expressions)

    @property
    def sub_type(self):
        return self._sub_type

    @property
    def is_embedded(self):
        return self._is_embedded

    @property
    def destination_folder_id(self):
        return self._destination_folder_id

    @property
    def expressions(self):
        """List of `FactExpressions` defined in `Fact` object."""
        return self._expressions

    @property
    def entry_level(self):
        """Natural level (a set of attributes) of the fact."""
        return self._entry_level



from mstrio.connection import Connection
from mstrio.users_and_groups.user import User
from mstrio.utils.entity import auto_match_args_entity
from mstrio.utils.helper import Dictable


class CertifiedInfo(Dictable):
    """Certification status, time of certification and information
    about the certifier (currently only for document and report).

    Attributes:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`.
        certified: Specifies whether the object is trusted,
            as determined by the standards set by the certifier
        date_certified: time when the object was certified,
            "yyyy-MM-dd HH:mm:ss" in UTC
        certifier: information about the entity certifying
            the object, User object
    """

    def __init__(
        self,
        connection: Connection,
        certified: bool,
        date_certified: str | None = None,
        certifier: dict | None = None,
    ):
        self._connection = connection
        self._certified = certified
        self._date_certified = date_certified
        self._certifier = (
            User.from_dict(certifier, self._connection) if certifier else None
        )

    def __str__(self):
        if not self.certified:
            return 'Object is not certified.'
        return f"Object certified on {self.date_certified} by {self._certifier}"

    def __repr__(self):
        param_value_dict = auto_match_args_entity(
            self.__init__, self, exclude=['self'], include_defaults=False
        )
        params_list = []
        for param, value in param_value_dict.items():
            if param == "connection" and isinstance(value, Connection):
                params_list.append("connection")
            else:
                params_list.append(f"{param}={repr(value)}")
        formatted_params = ", ".join(params_list)
        return f"{self.__class__.__name__}({formatted_params})"

    @property
    def connection(self) -> Connection:
        return self._connection

    @property
    def certified(self) -> bool | None:
        return self._certified

    @property
    def date_certified(self) -> str | None:
        return self._date_certified

    @property
    def certifier(self) -> User | None:
        return self._certifier



from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error loading dataset {id} Check dataset ID")
def dataset_definition(connection, id, fields=None, whitelist=None):
    """Get the definition of a dataset.

    Args:
        connection (object): MicroStrategy connection object returned by
            connection.Connection()`.
        id (str): Identifier of a pre-existing dataset. Used when
            updating a pre-existing dataset.
        fields(list, optional): Specifies object types to be returned. Possible
            values include tables, columns, attributes, and metrics. If no value
            is set, attributes and metrics are returned.
        whitelist(list): list of errors for which we skip printing error
            messages

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    connection._validate_project_selected()
    return connection.get(endpoint=f'/api/datasets/{id}', params={'fields': fields})


@ErrorHandler(err_msg="Error creating new dataset model.")
def create_dataset(connection, body):
    """Create a single-table dataset from external data uploaded to the
    MicroStrategy Intelligence Server.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        body (str): JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    connection._validate_project_selected()
    return connection.post(endpoint='/api/datasets', json=body)


@ErrorHandler(err_msg="Error updating dataset with ID {id}")
def update_dataset(connection, id, table_name, update_policy, body):
    """Update a single-table dataset with external data uploaded to the
    MicroStrategy Intelligence Server.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Identifier of a pre-existing dataset. Used when
            updating a pre-existing dataset.
        table_id (str): Identifier of the table to update within the
            MicroStrategy dataset.
        update_policy (str):  Update operation type: 'Add' (inserts new, unique
            rows), 'Update' (updates data in existing rows and columns),
            'Upsert' (updates existing data and inserts new rows), 'Replace'
            (similar to truncate, replaces the existing data with new data).
        body (str): JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.patch(
        endpoint=f'/api/datasets/{id}/tables/{table_name}',
        headers={'updatePolicy': update_policy},
        json=body,
    )


@ErrorHandler(err_msg="Error deleting dataset with ID {id}")
def delete_dataset(connection, id):
    """Delete a dataset previously created using the REST API.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Identifier of a pre-existing dataset. Used when
            updating a pre-existing dataset.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.delete(endpoint=f'/api/objects/{id}?type=3')


@ErrorHandler(err_msg="Error creating new dataset model.")
def create_multitable_dataset(connection, body):
    """Create the definition of a multi-table dataset.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        body (dict): JSON-formatted payload containing the body of the request.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    connection._validate_project_selected()
    return connection.post(endpoint='/api/datasets/models', json=body)


@ErrorHandler(err_msg="Error creating new data upload session.")
def upload_session(connection, id, body):
    """Create a multi-table dataset upload session.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Identifier of a pre-existing dataset. Used when
            updating a pre-existing dataset.
        body (dict): JSON-formatted payload containing the body of the request.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    connection._validate_project_selected()
    return connection.post(endpoint=f'/api/datasets/{id}/uploadSessions', json=body)


@ErrorHandler(err_msg="Error uploading data to dataset {id}")
def upload(connection, id, session_id, body, throw_error=False):
    """Upload data to a multi-table dataset.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Identifier of a pre-existing dataset. Used when
            updating a pre-existing dataset.
        session_id (str): Identifier of the server session used for collecting
            uploaded data.
        body (dict): JSON-formatted payload containing the body of the request.
        throw_error (bool): Flag indicates if the error should be thrown

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    connection._validate_project_selected()
    return connection.put(
        endpoint=f'/api/datasets/{id}/uploadSessions/{session_id}',
        json=body,
    )


@ErrorHandler(
    err_msg="Error publishing uploaded data for dataset with ID {id} Cancelling "
    "publication."
)
def publish(connection, id, session_id, throw_error=False):
    """Publish a multi-table dataset.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Identifier of a pre-existing dataset. Used when
            updating a pre-existing dataset.
        session_id (str): Identifier for the server session used for collecting
            uploaded data.
        throw_error (bool): Flag indicates if the error should be thrown

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    connection._validate_project_selected()
    endpoint = f'/api/datasets/{id}/uploadSessions/{session_id}/publish'
    return connection.post(endpoint=endpoint)


@ErrorHandler(err_msg="Failed to check dataset {id} publication status.")
def publish_status(connection, id, session_id):
    """Get multi-table dataset publication status.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Identifier of a pre-existing dataset. Used when
            updating a pre-existing dataset.
        session_id (str): Identifier for the server session used for collecting
            uploaded data.

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """

    endpoint = f'/api/datasets/{id}/uploadSessions/{session_id}/publishStatus'
    return connection.get(endpoint=endpoint)


@ErrorHandler(err_msg="Failed to cancel the publication of dataset with ID {id}")
def publish_cancel(connection, id, session_id, throw_error=False):
    """Delete a multi-table dataset upload session and cancel publication.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Identifier of a pre-existing dataset. Used when
            updating a pre-existing dataset.
        session_id (str): Identifier for the server session used for collecting
            uploaded data.

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    endpoint = f'/api/datasets/{id}/uploadSessions/{session_id}'
    return connection.delete(endpoint=endpoint)



from mstrio.api import usergroups as usergroups_api
from mstrio.api import users as users_api
from mstrio.connection import Connection
from mstrio.utils.helper import deprecation_warning, fetch_objects, fetch_objects_async
from mstrio.utils.version_helper import method_version_handler


def get(connection: Connection, id: str):
    """Get user by a specified ID.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the user

    Returns:
        dict representing user object
    """
    return users_api.get_user_info(connection=connection, id=id).json()


def get_addresses(connection: Connection, id: str):
    """Get addresses for a specified user.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the user

    Returns:
        dict representing user addresses
    """
    return users_api.get_addresses_v2(connection=connection, id=id).json()


def get_security_roles(connection: Connection, id: str):
    """Get security roles for a specified user.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the user

    Returns:
        dict representing user security roles
    """
    return users_api.get_user_security_roles(connection=connection, id=id).json()


def get_settings(connection: Connection, id: str):
    """Get settings for a specified user.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the user

    Returns:
        dict representing user security roles
    """
    return users_api.get_settings(connection=connection, id=id).json()


def get_privileges(connection: Connection, id: str):
    """Get privileges for a specified user.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the user

    Returns:
        dict representing user privileges
    """
    return users_api.get_user_privileges(connection=connection, id=id).json()


def update(connection: Connection, id: str, body: dict):
    """Update info for a specified user.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the user
        body: body of the request

    Returns:
        dict representing user object
    """
    return users_api.update_user_info(connection=connection, id=id, body=body).json()


def create(connection: Connection, body: dict, username: str):
    """Create a user.

    Args:
        connection: MicroStrategy REST API connection object
        body: body of the request
        username: name of the user, for error purposes only

    Returns:
        dict representing user object
    """
    return users_api.create_user(
        connection=connection, body=body, username=username
    ).json()


def get_all(
    connection: Connection,
    limit: int,
    msg: str,
    name_begins: str,
    abbreviation_begins: str,
    filters: dict,
):
    """Get list of users.

    Args:
        connection: MicroStrategy REST API connection object
        limit: limit of users to list
        msg: optional error message,
        name_begins: optional filter for name beginning with
        abbreviation_begins: optional filter for abbreviation beginning with
        filters: filters

    Returns:
        list of dicts representing users
    """
    if filters.get('initials'):
        deprecation_warning(
            deprecated="possibility of providing 'initials' as a filter",
            new="New options to filter on 'full_name' and 'enabled' fields instead",
            version="11.3.13.101",
            module=False,
        )
        return fetch_objects_async(
            connection=connection,
            api=users_api.get_users_info,
            async_api=users_api.get_users_info_async,
            limit=limit,
            chunk_size=1000,
            error_msg=msg,
            name_begins=name_begins,
            abbreviation_begins=abbreviation_begins,
            filters=filters,
        )
    if filters.get('name') or name_begins:
        name_begins_filter = ('starts', name_begins) if name_begins else None

        filters['name'] = filters.get('name') or name_begins_filter

    if filters.get('abbreviation') or abbreviation_begins:
        abbreviation_begins_filter = (
            ('starts', abbreviation_begins) if abbreviation_begins else None
        )
        filters['abbreviation'] = (
            filters.get('abbreviation') or abbreviation_begins_filter
        )
    # Getting information from members of 'Everyone' user group
    return fetch_objects(
        connection=connection,
        api=usergroups_api.get_members,
        limit=limit,
        error_msg=msg,
        filters=filters,
        id='C82C6B1011D2894CC0009D9F29718E4F',
    )


def create_address(connection: Connection, id: str, body: dict):
    """Create an email user address.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the user
        body: body of the request

    Returns:
    created address dictionary
    """
    return users_api.create_address(connection=connection, id=id, body=body).json()


def create_address_v2(connection: Connection, id: str, body: dict):
    """Create a non-email user address.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the user
        body: body of the request

    Returns:
        dict representing user addresses
    """
    return users_api.create_address_v2(connection=connection, id=id, body=body).json()


def update_address(connection: Connection, id: str, address_id: str, body: dict):
    """Update a user address.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the user
        address_id: ID of the address
        body: body of the request

    Returns:
        True if successfully updated, False otherwise
    """
    return users_api.update_address_v2(
        connection=connection, id=id, address_id=address_id, body=body
    ).ok


def delete_address(connection: Connection, id: str, address_id: str):
    """Delete a user address.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the user
        address_id: ID of the address

    Returns:
        True if successfully deleted, False otherwise
    """
    return users_api.delete_address(
        connection=connection, id=id, address_id=address_id
    ).ok


def get_security_filters(connection: Connection, id: str, projects: str | list[str]):
    """Get security filters for a user.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the user
        projects: IDs of the projects

    Returns:
        list of user security filters
    """
    return users_api.get_security_filters(
        connection=connection, id=id, projects=projects
    ).json()


@method_version_handler('11.3.0700')
def update_user_settings(connection: Connection, id: str, body: str):
    """Update user settings.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the user

    Returns:
        True if successfully updated, False otherwise
    """
    response = users_api.update_user_settings(connection=connection, id=id, json=body)
    if response.ok:
        # Body is returned as endpoint returns 204 with empty response, which
        # would fail when using response.json() .
        # In order to alter user correctly, we return body to reflect changes
        return body
    return response



"""mstrio: Simple and Secure Access to MicroStrategy Data

Mstrio provides a high-level interface for Python and is designed
to give data scientists, developers, and administrators simple and secure
access to their MicroStrategy environment.

It wraps MicroStrategy REST APIs into simple workflows, allowing users
to fetch data from cubes and reports, create new datasets, add new data
to existing datasets, and manage Users/User Groups, Servers, Projects,
and more. Since it enforces MicroStrategy’s user and object security model,
you don’t need to worry about setting up separate security rules.

With mstrio-py for data science, it’s easy to integrate cross-departmental,
trustworthy business data in machine learning workflows and enable
decision-makers to take action on predictive insights in MicroStrategy Reports,
Dashboards, HyperIntelligence Cards, and customized, embedded analytical
applications.

With mstrio-py for system administration, it’s easy to minimize costs by
automating critical, time-consuming administrative tasks, even enabling
administrators to leverage the power of Python to address complex
administrative workflows for maintaining a MicroStrategy environment.
"""

__title__ = "mstrio-py"
__version__ = "11.4.3.101"  # NOSONAR
__license__ = "Apache License 2.0"
__description__ = "Python interface for the MicroStrategy REST API"
__author__ = "MicroStrategy"
__author_email__ = "bkaczynski@microstrategy.com"
__build__ = "11.4.0400.00001"



import fnmatch
import logging
from collections.abc import Callable, Iterable
from enum import Enum
from typing import TYPE_CHECKING, Any, TypeVar, Union

if TYPE_CHECKING:
    from mstrio.utils.entity import EntityBase

    SupportedExpression = Union[
        list, str, dict, int, float, bool, EntityBase, Enum, tuple
    ]

logger = logging.getLogger(__name__)

SupportedExpression = Union[list, str, dict, int, float, bool, Enum]
KT = TypeVar("KT")
VT = TypeVar("VT")

EQUAL = '='
SMALLER_EQUAL = '<='
LARGER_EQUAL = '>='
NOT_EQUAL = '!='
NOT = '!'
SMALLER = '<'
LARGER = '>'
IS = 'is'
IN = 'in'
DICT_COMPARE = 'dict'
ENTITY_COMPARE = 'entity'
STARTS = 'starts'


def check_valid_param(dict_object: dict[KT, VT], params: Iterable) -> None:
    """Check if filter parameters can be used with given dict."""
    # all keys from dict that are supported for filtering
    allowed = list(
        filter(
            lambda el: not isinstance(dict_object[el], (list, tuple, set)),
            dict_object.keys(),
        )
    )

    # check filter param is in the allowed
    for param in params:
        if param not in allowed:
            raise KeyError(
                f"The filter parameter '{param}' is not valid. "
                f"Please filter by one of: {allowed}"
            )


def parse_filter_expression(param: str, expression: SupportedExpression) -> tuple:
    """Parse the filter expression only once"""
    from mstrio.utils.entity import EntityBase

    op = None

    # unpack enum value
    if isinstance(expression, Enum):
        expression = expression.value

    if isinstance(expression, list):
        op = IN
        filter_value = expression
    elif isinstance(expression, str):
        # extract the operation from the expression if it exists
        if expression and expression[0] in [SMALLER, LARGER, NOT, EQUAL, STARTS]:
            op = expression[0]
            if expression[0] in [SMALLER, LARGER, NOT] and expression[1] == EQUAL:
                op += expression[1]
        filter_value = expression[len(op) :] if op is not None else expression

        if filter_value.lower() == 'true':  # Support string bool values
            filter_value = True
            op = IS
        elif filter_value.lower() == 'false':
            filter_value = False
            op = IS
        op = op or EQUAL
    elif isinstance(expression, bool):
        op = IS
        filter_value = expression
    elif isinstance(expression, (int, float)):
        op = EQUAL
        filter_value = expression
    elif isinstance(expression, dict):
        op = DICT_COMPARE
        filter_value = expression
    elif isinstance(expression, EntityBase):
        op = ENTITY_COMPARE
        filter_value = expression
    elif isinstance(expression, tuple):
        op, filter_value = expression
    else:
        raise TypeError(
            f"'{param}' filter value must be either a string, "
            "bool, int, float, dict, tuple or list"
        )
    return (op, filter_value)


def make_dict_filter(param: str, op: str, filter_value: Any) -> Callable:
    """Return a filter function that takes a dictionary object as parameter.

    Once evaluated it return bool value indicating if given parameter-
    expression is True or False. This function can be used in the
    filter() method.
    """

    def cast_filter_value_if_needed(value_type: type) -> Any:
        # entity -> Entity will take care of ValueType errors -> don't cast
        # dict -> type must match exactly -> don't cast
        # str -> type must match exactly -> cast
        # int -> type must match exactly to int or float -> cast
        # float -> type must match exactly to int or float -> cast
        # bool -> type must match exactly to bool -> cast
        # enum -> could be anything as value is extracted

        if (op == ENTITY_COMPARE) or (
            isinstance(filter_value, value_type) and op != IN
        ):
            return filter_value
        elif not isinstance(filter_value, value_type):
            # check if all other types are not equal to value_type and cast
            if op in (DICT_COMPARE, IS):
                # do not cast to bool or to dict
                raise TypeError(
                    f"'{param}' needs to be compatible with type {value_type}."
                )

            try:
                # cast filter_value to dict element type
                if op == IN:
                    return [value_type(val) for val in filter_value]

                else:
                    # cast filter_value to int, float, str as in dict
                    return value_type(filter_value)
            except ValueError as e:
                logger.error(f"'{param}' filter value is incorrect.")
                raise e

    def my_filter(dict_object: dict[KT, VT]) -> bool:
        """This function will be executed when passed into filter() function"""
        # extract actual value from dict
        if param not in dict_object:
            return False

        value = dict_object[param]
        value_type = type(value)
        filter_value = cast_filter_value_if_needed(value_type)

        if op in (EQUAL, ENTITY_COMPARE):
            return value == filter_value
        elif op in (NOT_EQUAL, NOT):
            return value != filter_value
        elif op == IS:
            return value is filter_value
        elif op == LARGER:
            return value > filter_value
        elif op == SMALLER:
            return value < filter_value
        elif op == LARGER_EQUAL:
            return value >= filter_value
        elif op == SMALLER_EQUAL:
            return value <= filter_value
        elif op == IN:
            return value in filter_value
        elif op == STARTS:
            return fnmatch.fnmatch(value, f'{filter_value}*')
        elif op == DICT_COMPARE:
            return all(
                (
                    value[k] == v if k in value else False
                    for k, v in filter_value.items()
                )
            )

    return my_filter


def filter_list_of_dicts(
    list_of_dicts: list[dict[KT, VT]], **filters: dict[str, SupportedExpression]
) -> list[dict[KT, VT]]:
    """Filter a list of dicts by providing one or more key-value pair filters.

    Args:
        list_of_dicts: list of dicts that will be filtered
        **kwargs: Supports filtering by list, str, dict, int, float, bool,
            EntityBase, Enum as filter value.
            Support simple logical operators like: '<,>,<=,>=,!' if value
            is str.

    Examples:
        >>>filter_list_of_dicts(l, val=[0, 1, 2], val2=">2", val3=False)
        >>>filter_list_of_dicts(l, val=2, val2="!2", val3=SomeEnum(1))
        >>>filter_list_of_dicts(l, val={"id":"123", "name":"xxx"})
        >>>filter_list_of_dicts(l, val=User(conn, id="123"))
    """
    # check dict keys include the filter (only first element due to performance)
    if list_of_dicts:
        model_dict = list_of_dicts[0]
        check_valid_param(model_dict, filters.keys())

        for param, expression in filters.items():
            op, filter_value = parse_filter_expression(param, expression)
            filter_function = make_dict_filter(param, op, filter_value)
            list_of_dicts = list(filter(filter_function, list_of_dicts))

    return list_of_dicts



from mstrio.connection import get_connection
from mstrio.project_objects import Report, list_reports

# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
connection = get_connection(workstationData, project_name='MicroStrategy Tutorial')

# List available Reports (limited to 10)
sample_reports = list_reports(connection, limit=10)
for report in sample_reports:
    print(report)

# Check available attributes and metrics of a Report
sample_report_id = '759C0B5340E0AD6FAEDFF19D4DBC3488'
sample_report = Report(connection, id=sample_report_id)
attributes = sample_report.attributes
metrics = sample_report.metrics
print(f"Attributes: {attributes}\nMetrics: {metrics}")

# Choose attributes and metrics
category_attribute = attributes[0].get('id')
subcategory_attribute = attributes[1].get('id')

cost_metric = metrics[0].get('id')
profit_metric = metrics[1].get('id')

# Filter which attributes and metrics to use in a dataframe
sample_report.apply_filters(
    attributes=[category_attribute, subcategory_attribute],
    metrics=[cost_metric, profit_metric],
)

# Create a dataframe from a Report
dataframe = sample_report.to_dataframe()
print(dataframe.head())



"""This is the demo script to show how to manage intelligent (OLAP) cubes. 

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to ease
its usage.
"""

from mstrio.project_objects import list_olap_cubes, OlapCube
from mstrio.connection import get_connection
from mstrio.project_objects.datasets.helpers import (
    AttributeTemplateUnit,
    MetricTemplateUnit,
    AttributeFormReference,
    MetricElement,
    Template
)

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert project to connect to here

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Define a variable which can be later used in a script
CUBE_ID = $cube_id  # ID for OLAP Cube object lookup here

# List OLAP Cubes with different conditions
list_of_all_cubes = list_olap_cubes(connection=conn)
print(list_of_all_cubes)
list_of_all_cubes_as_dicts = list_olap_cubes(connection=conn, to_dictionary=True)
print(list_of_all_cubes_as_dicts)
list_of_one_olap_cube = list_olap_cubes(connection=conn, id=CUBE_ID)
print(list_of_one_olap_cube)

# Get OLAP Cube by its ID
olap_cube = OlapCube(conn, CUBE_ID)

# List available attributes, metrics and attribute forms which can be used for
# creation of a new OLAP Cube
OlapCube.available_metrics(conn)
OlapCube.available_attributes(conn)

# Define variables which can be later used in a script
ATTRIBUTE_ID = $attribute_id            # Insert ID of an attribute here
ATTRIBUTE_NAME = $attribute_name        # Insert name of an attribute here
METRIC_ID = $metric_id                  # Insert ID of a metric here
METRIC_NAME = $metric_name              # Insert name of a metric here
CUBE_NAME = $cube_name                  # Insert name of a cube here
CUBE_DESCRIPTION = $cube_description    # Insert description for a cube here
CUBE_FOLDER_ID = $cube_folder_id        # Insert ID for a cube folder here
ATTRIBUTE = {'id': ATTRIBUTE_ID, 'name': ATTRIBUTE_NAME, 'type': 'attribute'}
METRIC = {'id': METRIC_ID, 'name': METRIC_NAME, 'type': 'metric'}

# Create new OLAP Cube with attributes and metrics
attributes = [ATTRIBUTE]
metrics = [METRIC]
new_olap_cube_from_attributes_and_metrics = OlapCube.create(
    conn,
    name=CUBE_NAME,
    description=CUBE_DESCRIPTION,
    folder_id=CUBE_FOLDER_ID,
    attributes=attributes,
    metrics=metrics
)

# Define variables which can be later used in a script
TEMPLATE = $template                        # Insert template here
FILTER_EXPRESSION = $filter_expression      # Insert filter here
CUBE_OPTIONS = $cube_options                # Insert cube options here
ADVANCED_PROPERTIES = $advanced_properties  # Insert advanced properties here
TIME_BASED_SETTINGS = $time_based_settings  # Insert time based settings here

# Create an OLAP Cube with template
new_olap_cube = OlapCube.create(
    connection=conn,
    name=CUBE_NAME,
    folder_id=CUBE_FOLDER_ID,
    description=CUBE_DESCRIPTION,
    template=TEMPLATE,
    filter=FILTER_EXPRESSION,
    options=CUBE_OPTIONS,
    advanced_properties=ADVANCED_PROPERTIES,
    time_based_settings=TIME_BASED_SETTINGS,
)

# Define variables which can be later used in a script
NEW_CUBE_NAME = $new_cube_name                      # Insert new name of cube here
NEW_CUBE_DESCRIPTION = $new_cube_description        # Insert new description for cube here
NEW_CUBE_ABBREVIATION = $new_cube_abbreviation      # Insert new abbreviation for cube here
NEW_TEMPLATE = $new_template                        # Insert new template here
NEW_FILTER_EXPRESSION = $new_filter_expression      # Insert new filter here
NEW_CUBE_OPTIONS = $new_cube_options                # Insert new cube options here
NEW_TIME_BASED_SETTINGS = $new_time_based_settings  # Insert new time based settings here

# Alter an OLAP Cube
new_olap_cube.alter(
    name=NEW_CUBE_NAME,
    description=NEW_CUBE_DESCRIPTION,
    abbreviation=NEW_CUBE_ABBREVIATION,
    template=NEW_TEMPLATE,
    filter=NEW_FILTER_EXPRESSION,
    options=NEW_CUBE_OPTIONS,
    time_based_settings=NEW_TIME_BASED_SETTINGS,
)

# Define variables which can be later used in a script
NEW_ATTRIBUTE_ID = $new_attribute_id
NEW_ATTRIBUTE_NAME = $new_attribute_name
NEW_METRIC_ID = $new_metric_id
NEW_METRIC_NAME = $new_metric_name
NEW_ATTRIBUTE = {'id': NEW_ATTRIBUTE_ID, 'name': NEW_ATTRIBUTE_NAME, 'type': 'attribute'}
NEW_METRIC = {'id': NEW_METRIC_ID, 'name': NEW_METRIC_NAME, 'type': 'metric'}
# Update attributes and metrics of a newly created OLAP Cube. When cube is not
# published yet, then it is possible to add/remove metrics or attributes.
# When cube is published it is possible to only change order of attributes or
# metrics which have already been added to cube.
attributes = [ATTRIBUTE, NEW_ATTRIBUTE]
metrics = [METRIC, NEW_METRIC]
new_olap_cube_from_attributes_and_metrics.update(attributes=attributes, metrics=metrics)
new_olap_cube.update(attributes=attributes, metrics=metrics)

# Please note that if AttributeTempateUnit.forms are empty and equals to None
# default attribute forms will be used to create OLAP cube and string indicating that
# will be returned instead of forms from method OlapCube.list_attribute_forms().
# Also if AttributeTempateUnit.forms are specified, OlapCube.to_dataframe() will
# return corresponding to selected attribute forms. To add new form for an attribute
# in OLAP Cube and to see reflected changes in generated dataframe use OlapCube.alter
# with modified template with newly added form.
# Define variables which can be later used in a script
NEW_ATTRIBUTE_FORM_ID = $new_attribute_form_id      # Insert new attribute form ID here
TEMPLATE_WITH_NEW_FORM = Template(
    rows=[
        AttributeTemplateUnit(
            id='8D679D4211D3E4981000E787EC6DE8A4',
            name='Item',
            alias=None,
            forms=[
                AttributeFormReference(id='45C11FA478E745FEA08D781CEA190FE5'),
                AttributeFormReference(id='CCFBE2A5EADB4F50941FB879CCF1721C'),
            ],
        ),
        MetricTemplateUnit(
            elements=[
                MetricElement(
                    id='9F9B1E724E4E242755BE10AFABA290F4',
                    name='Web Sales',
                    alias=None,
                ),
            ]
        ),
    ],
    columns=[
        AttributeTemplateUnit(
            id='8D679D5011D3E4981000E787EC6DE8A4',
            name='Supplier',
            alias=None,
            forms=[
                AttributeFormReference(id='45C11FA478E745FEA08D781CEA190FE5'),
                AttributeFormReference(id='CCFBE2A5EADB4F50941FB879CCF1721C'),
                AttributeFormReference(id=NEW_ATTRIBUTE_FORM_ID),
            ],
        ),
        MetricTemplateUnit(
            elements=[
                MetricElement(
                    id='971752994EFBAE6CF37A26A03FDA9813',
                    name='Average Revenue',
                    alias=None,
                ),
            ]
        ),
        AttributeTemplateUnit(
            id='8D679D4B11D3E4981000E787EC6DE8A4',
            name='Region',
        ),
    ],
    page_by=[],
)

# To add new attribute form to OLAP cube
new_olap_cube.alter(
    template=TEMPLATE_WITH_NEW_FORM,
)

# Set new partition attribute for OLAP Cube by ID
new_olap_cube.set_partition_attribute(ATTRIBUTE_ID)
# Remove partition attribute for OLAP Cube
new_olap_cube.remove_partition_attribute()
# Listing attribute forms for OLAP Cube
attribute_forms = new_olap_cube.list_attribute_forms()

# Publish newly created and updated OLAP Cube
new_olap_cube_from_attributes_and_metrics.publish()
new_olap_cube.publish()

# Refresh and show status of an OLAP Cube
new_olap_cube_from_attributes_and_metrics.refresh_status()
new_olap_cube.refresh_status()
new_olap_cube_from_attributes_and_metrics.show_status()
new_olap_cube.show_status()

# Export sql view of an OLAP Cube
sql_view = new_olap_cube_from_attributes_and_metrics.export_sql_view()
sql_view = new_olap_cube.export_sql_view()

# Unpublish an OLAP Cube
new_olap_cube_from_attributes_and_metrics.unpublish(force=True)
new_olap_cube.unpublish(force=True)

# Delete an OLAP Cube. When `force` argument is set to `False` (default value)
# then deletion must be approved.
new_olap_cube_from_attributes_and_metrics.delete(force=True)
new_olap_cube.delete(force=True)



from itertools import chain

import numpy as np
import pandas as pd

from mstrio.utils.helper import exception_handler


class Parser:
    """Converts JSON-formatted cube and report data into a tabular
    structure."""

    AF_COL_SEP = "@"  # attribute form column label separator; commonly "@"
    chunk_size = None
    total_rows = None

    def __init__(self, response, parse_cube=True):
        self.parse_cube = parse_cube
        # row-level metric data
        self._metric_values_raw = []

        # extract column headers and names
        self._metric_col_names = self.__extract_metric_col_names(response=response)

        # attribute objects
        # extract attribute names
        self._attribute_names = self.__extract_attribute_names(response=response)

        # extract attribute form names
        self._attribute_elem_form_names = self.__extract_attribute_form_names(
            response=response
        )

        # parse attribute column names including attribute form names
        # into final column names
        self._attribute_col_names = self.__get_attribute_col_names()

        self.__extract_paging_info(response)

        # attribute data
        self._mapped_attributes = np.zeros(
            (0, len(self._attribute_col_names)), dtype=object
        )

    def parse(self, response):
        """
        Args:
            response: JSON-formatted content of API response.
        """
        if self.total_rows > 0:
            # extract attribute values into numpy 2D array if attributes exist
            # in the response
            if self._attribute_names:
                self._mapped_attributes = np.vstack(
                    (self._mapped_attributes, self.__map_attributes(response=response))
                )

            # extract metric values if metrics exist in the response
            if self._metric_col_names:
                self._metric_values_raw.extend(
                    self.__extract_metric_values(response=response)
                )

    def __to_dataframe(self):
        # create attribute data frame, then re-map integer array with
        # corresponding attribute element values
        attribute_df = pd.DataFrame(
            data=self._mapped_attributes, columns=self._attribute_col_names
        )

        # create metric values data frame
        metric_df = pd.DataFrame(
            data=self._metric_values_raw, columns=self._metric_col_names
        )

        return pd.concat([attribute_df, metric_df], axis=1)

    def __map_attributes(self, response):
        label_map = self.__create_attribute_element_map(response=response)
        row_index = self.__extract_attribute_element_row_index(response=response)
        row_index_array = np.array(row_index)
        _, columns = row_index_array.shape

        # create attribute form name to attribute form index mapping
        vfunc = np.vectorize(
            lambda attribute_indexes, columns: label_map[columns][attribute_indexes]
        )

        return vfunc(row_index_array, range(columns))

    def __create_attribute_element_map(self, response):
        """Create a map of type nested list for attribute element labels.

        The index of the list corresponds to attribute element row index from
        the grid headers. The map is used to map the integer-based grid
        header values to the real attribute element labels."""

        # extract attribute form and attribute elements labels
        rows = response["definition"]["grid"]["rows"]
        form_values_rows = [
            [el['formValues'] for el in row['elements']] for row in rows
        ]

        def replicate_form_values(form_values):
            """Replicate attribute element values for total, count, etc. to fill
            the lists to correct size. I-Server only sends them once per
            attribute"""
            if self.parse_cube:
                return

            for i, attr in enumerate(self._attribute_elem_form_names):
                required_len = len(attr)

                for r in form_values[i]:
                    if len(r) < required_len:
                        r.extend(r * (required_len - 1))

        def separate(form_values):
            """Format into correct list structure. This function separates the
            nested lists into a list of list where one element corresponds to an
            attribute column"""
            final_list = []
            try:
                for attr in form_values:
                    row = len(attr[0])
                    col = len(attr)
                    final_list.extend(
                        np.array(attr).reshape(col, row).transpose().tolist()
                    )
                return final_list
            except IndexError:
                msg = (
                    "Missing attribute elements, please check if attribute elements IDs"
                    " are valid and if they exist in report."
                )
                exception_handler(msg, IndexError)

        replicate_form_values(form_values_rows)
        ae_index_map = separate(form_values_rows)

        return ae_index_map

    def __extract_attribute_element_row_index(self, response):
        # extracts the attribute element row index from the headers
        return [
            list(
                chain.from_iterable(
                    [
                        [r for _ in f]
                        for r, f in zip(row, self._attribute_elem_form_names)
                    ]
                )
            )
            for row in response["data"]["headers"]["rows"]
        ]

    def __extract_paging_info(self, response):
        # extract paging info
        self.chunk_size = response["data"]["paging"]["limit"]
        self.total_rows = response["data"]["paging"]["total"]

    @staticmethod
    def __extract_metric_values(response):
        return response["data"]["metricValues"]["raw"]

    @staticmethod
    def __extract_metric_col_names(response):
        if response["definition"]["grid"]["columns"]:
            return [
                i['name']
                for i in response["definition"]["grid"]["columns"][-1]["elements"]
            ]
        else:
            return []

    @staticmethod
    def __extract_attribute_form_names(response):
        # extract attribute form names
        return [
            [form["name"] for form in attribute["forms"]]
            for attribute in response["definition"]["grid"]["rows"]
        ]

    @staticmethod
    def __extract_attribute_names(response):
        return [
            attribute["name"] for attribute in response["definition"]["grid"]["rows"]
        ]

    def __get_attribute_col_names(self):
        # extract and format attribute form column labels
        col_names = []
        for attr, forms in zip(self._attribute_names, self._attribute_elem_form_names):
            if len(forms) == 1:
                # if only one form, do not display the attribute form type
                # in the column headers
                col_names.append(attr)
            else:
                # otherwise concatenate the attribute name, separator,
                # and form type
                for form in forms:
                    col_names.append(attr + self.AF_COL_SEP + form)

        return col_names

    def has_multiform_attributes(self):
        return any(len(item) > 1 for item in self._attribute_elem_form_names)

    @property
    def dataframe(self):
        return self.__to_dataframe()



"""
This demo script shows how to connect to MicroStrategy Environment. This script
will not work without replacing parameters with real values. Its basic goal is
to present what can be done with this module and to ease its usage.
"""

from mstrio.connection import get_connection, Connection

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert project name here

# The Connection object manages your connection to MicroStrategy. Connect to
# your MicroStrategy environment by providing the URL to the MicroStrategy REST
# API server, your username, password and the ID of the Project to connect to.
# When a Connection object is created the user will be automatically logged-in.
# Connection object automatically renews the connection or reconnects,
# if session becomes inactive. Reconnection doesn't work if authenticated with
# identity token.
conn = get_connection(workstationData, project_name=PROJECT_NAME)
# The URL for the REST API server typically follows this format:
# https://your-microstrategy-server.com/MicroStrategyLibrary/api

# Validate that the REST API server is running by accessing
# https://your-microstrategy-server.com/MicroStrategyLibrary/api-docs in your
# web browser.

# To manage the connection the following methods are made available:
conn.connect()
conn.close()
conn.status()

# Define variables which can be later used in a script
BASE_URL = $base_url  # Insert URL for your env here
MSTR_USERNAME = $mstr_username  # Insert your env username here
MSTR_PASSWORD = $mstr_password  # insert your mstr password here
PROJECT_ID = $project_id  # Insert you project ID here

# Authentication Methods
# Currently, supported authentication modes are Standard (the default) and LDAP.
# To use LDAP, add login_mode=16 when creating your Connection object:
conn = Connection(BASE_URL, MSTR_USERNAME, MSTR_PASSWORD, project_id=PROJECT_ID, login_mode=16)

# Define a variable which can be later used in a script
IDENTITY_TOKEN = $identity_token  # Insert your identity token here

# Optionally, the Connection object can be created by passing the IDENTITY_TOKEN
# parameter, which will create a delegated session. The identity token can be
# obtained by sending a request to MicroStrategy REST API /auth/identityToken
# endpoint.
conn = Connection(BASE_URL, identity_token=IDENTITY_TOKEN, project_id=PROJECT_ID)

# SSL Self - signed Certificate
# By default, SSL certificates are validated with each API request.
# To turn this off, use ssl_verify flag:
conn = Connection(BASE_URL, MSTR_USERNAME, MSTR_PASSWORD, project_id=PROJECT_ID, ssl_verify=False)

# If you are using an SSL with a self-signed certificate you will need to
# perform an additional step to configure your connection. There are 2 ways to
# set it up:

# The easiest way to configure the SSL is to move your certificate file to your
# working directory. Just make sure the ssl_verify parameter is set to True when
# creating the Connection object in mstrio - py(it is True by default):
conn = Connection(BASE_URL, MSTR_USERNAME, MSTR_PASSWORD, project_id=PROJECT_ID, ssl_verify=True)

# Define a variable which can be later used in a script
CERTIFICATE_PATH = $certificate_path  # Insert your certificate path here

# The second way is to pass the certificate_path parameter to your connection
# object in mstrio. It has to be the absolute path to your certificate file:
conn = Connection(
    BASE_URL,
    MSTR_USERNAME,
    MSTR_PASSWORD,
    project_id=PROJECT_ID,
    certificate_path=CERTIFICATE_PATH
)

# Define variables which can be later used in a script
PROXIES_HTTP_VALUE = $proxies_http_value  # f.e. 'foo.bar:3128'
PROXIES_HTTPS_KEY = $proxies_https_key  # f.e. 'https://host.name'
PROXIES_HTTPS_VALUE = $proxies_https_value # f.e. 'foo.bar:4012'
PROXIES = {'http': PROXIES_HTTP_VALUE, PROXIES_HTTPS_KEY: PROXIES_HTTPS_VALUE }
# f.e. 'https://<username>:<password>@<ip_address>:<port>/'
PROXIES_HTTP_VALUE_USERNAME_AND_PASSWORD = $proxies_http_value_username_and_password

# Proxy
# Optionally, proxy settings can be set for the MicroStrategy Connection object.
conn = Connection(BASE_URL, MSTR_USERNAME, MSTR_PASSWORD, project_id=PROJECT_ID, proxies=PROXIES)

# User can also specify username and password in proxies parameter to use HTTP
# Basic Auth:
proxies = {'http': PROXIES_HTTP_VALUE_USERNAME_AND_PASSWORD}
conn = Connection(BASE_URL, MSTR_USERNAME, MSTR_PASSWORD, project_id=PROJECT_ID, proxies=proxies)



from mstrio.api import objects as objects_api
from mstrio.connection import Connection
from mstrio.utils.helper import rename_dict_keys


def get_info(
    connection: Connection,
    id: str,
    object_type: int,
    project_id: str = None,
) -> dict:
    """Get information for a specific object in a specific project; if you do
    not specify a project ID, you get information for the object just in the
    non-project area.

    You identify the object with the object ID and object type. You specify
    the object type as a query parameter; possible values for object type are
    provided in EnumDSSXMLObjectTypes.

    Args:
        connection (Connection): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        object_type (int): One of EnumDSSXMLObjectTypes.
        project_id (str): ID of a project in which the object is located.

    Returns:
        dict
    """
    obj_info = objects_api.get_object_info(
        connection, id, object_type, project_id
    ).json()
    obj_info['hidden'] = obj_info.get('hidden', False)

    return obj_info


def update(connection, id: str, body: dict, object_type: int, project_id: str = None):
    """Get information for a specific object in a specific project; if you do
    not specify a project ID, you get information for the object in all
    projects.

    You identify the object with the object ID and object type. You specify
    the object type as a query parameter; possible values for object type are
    provided in EnumDSSXMLObjectTypes.

    Args:
        connection (Connection): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        body: (dict): body of the response
        object_type (int): One of EnumDSSXMLObjectTypes
        project_id (str): ID of a project in which the object is located.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    if owner := body.pop('owner', None):
        body['ownerId'] = owner

    if comments := body.pop('comments', None):
        body['longDescription'] = comments

    obj_info = objects_api.update_object(
        connection, id, body, object_type, project_id
    ).json()

    if (
        comments
        and not obj_info.get('comments')
        and not obj_info.get('longDescription')
    ):
        obj_info = get_info(connection, id, object_type, project_id)

    obj_info['hidden'] = obj_info.get('hidden', False)

    return rename_dict_keys(obj_info, {'longDescription': 'comments'})


def update_translations(
    connection: Connection,
    target_id: str,
    project_id: str,
    body: dict,
    object_type: int,
):
    """Create a translation based on provided body.

    Args:
        connection(Connection): MicroStrategy connection object returned by
            `connection.Connection()`
        target_id (str): ID of the Object the translation will be added for
        project_id (str): ID of the Project the Object is located in
        body (dict): json representation of the body for the request
        object_type (int): type of the Object the translation will be added for

    Returns:
        A dictionary representing list of translations for the Object.
    """
    return _prepare_rest_output_for_translations(
        objects_api.update_translations(
            connection=connection,
            project_id=project_id,
            id=target_id,
            object_type=object_type,
            body=body,
            fields=None,
        )
        .json()
        .get('localesAndTranslations')
    )


def get_translations(
    connection: Connection,
    project_id: str,
    target_id: str,
    object_type: int,
):
    """List translations of the Object.

    Args:
        connection(Connection): MicroStrategy connection object returned by
            `connection.Connection()`
        target_id (str): ID of the Object the translation will be listed for
        project_id (str): ID of the Project the Object is located in
        object_type (int): type of the Object the translation will be listed for

    Returns:
        A dictionary representing list of translations for the Object.
    """
    return _prepare_rest_output_for_translations(
        objects_api.get_translations(
            connection=connection,
            project_id=project_id,
            id=target_id,
            object_type=object_type,
            fields=None,
        )
        .json()
        .get('localesAndTranslations')
    )


def _prepare_rest_output_for_translations(rest_output):
    raw_list = []

    for target in rest_output:
        translations_list = []
        for translation in rest_output.get(target).get('translationValues'):
            translations_list.append(
                {
                    'value': rest_output.get(target)
                    .get('translationValues')
                    .get(translation)
                    .get('translation'),
                    'language_lcid': translation,
                }
            )
        raw_list.append(
            {
                'translation_target_id': target,
                'translation_target_name': rest_output.get(target).get(
                    'translationTargetName'
                ),
                'translation_values': translations_list,
            }
        )

    return raw_list



from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error getting document with ID {id}")
def get_document(connection, id, error_msg=None):
    """Get information for a document with document Id.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): Document ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/library/{id}'
    return connection.get(endpoint=endpoint)


@ErrorHandler(err_msg="Error unpublishing document with ID {id}")
def unpublish_document(connection, id, error_msg=None):
    """Unpublish a previously published document. This makes the document no
    longer available in the library of each user it was originally published
    to.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): Document ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/library/{id}'
    return connection.delete(endpoint=endpoint)


@ErrorHandler(err_msg="Error unpublishing document with ID {document_id}")
def unpublish_document_for_user(connection, document_id, user_id, error_msg=None):
    """Unpublish a previously published document. This makes the document no
    longer available in the library of each user specified in `user_id`

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        user_id (string): user ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    connection._validate_project_selected()
    endpoint = f'/api/library/{document_id}/recipients/{user_id}'
    return connection.delete(endpoint=endpoint)


@ErrorHandler(err_msg="Error getting library.")
def get_library(connection, error_msg=None):
    """Get the library for the authenticated user.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = '/api/library'
    return connection.get(endpoint=endpoint, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg="Error publishing document.")
def publish_document(connection, body, error_msg=None):
    """Publish a document to users or user groups in a specific project.

    Args:
        connection: MicroStrategy REST API connection object
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    connection._validate_project_selected()
    endpoint = '/api/library'
    return connection.post(endpoint=endpoint, json=body)



from mstrio.utils import helper


class Filter:
    err_msg_invalid = "Invalid object ID: '{}'"
    err_msg_duplicated = "Duplicate object ID: '{}'"

    def __init__(
        self,
        attributes: list,
        metrics: list,
        attr_elements: list | None = None,
        row_count_metrics: list | None = None,
        operator: str = 'In',
    ):
        self.attributes = {}
        for a in attributes:
            self.attributes[a["id"]] = {"name": a["name"]}

        self.metrics = {}
        for m in metrics:
            self.metrics[m["id"]] = {"name": m["name"]}

        self.attr_elems = {}
        if attr_elements is not None:
            self._populate_attr_elements(attr_elements)
        self.row_count_metrics = (
            [] if row_count_metrics is None else [i['id'] for i in row_count_metrics]
        )
        self.attr_selected = []
        self.metr_selected = []
        self.attr_elem_selected = []
        self.operator = operator

        # select all metrics and all attributes
        self._select([metric_id['id'] for metric_id in metrics])
        self._select([attribute_id['id'] for attribute_id in attributes])

    def _select(self, object_id):
        attr_form_object_id = None
        if isinstance(object_id, list):
            for i in set(object_id):
                self._select(object_id=i)
            return

        if len(object_id) > 32 and object_id[32] == ';':
            attr_form_object_id = [object_id[:32], object_id[33:]]
            object_id = attr_form_object_id[0]

        if self.__invalid(object_id):
            raise ValueError(self.err_msg_invalid.format(object_id))
        if self.__duplicated(object_id):
            helper.exception_handler(
                msg=self.err_msg_duplicated.format(object_id), exception_type=Warning
            )
        else:
            typ = self.__type(object_id)

            if typ == "attribute":
                self.attr_selected.append(attr_form_object_id or [object_id])

            if typ == "metric":
                self.metr_selected.append(object_id)

    def _select_attr_el(self, element_id):
        if isinstance(element_id, list):
            for i in element_id:
                self._select_attr_el(element_id=i)
        else:
            attribute_id = element_id.split(':')[0]
            if self.__invalid(attribute_id):
                raise ValueError(self.err_msg_invalid.format(element_id))

            if self.__duplicated(element_id):
                helper.exception_handler(
                    msg=self.err_msg_duplicated.format(element_id),
                    exception_type=Warning,
                )
            else:
                self.attr_elems[element_id] = {}
                self.attr_elems[element_id]["name"] = element_id
                self.attr_elems[element_id]["attribute_id"] = attribute_id
                self.attr_elem_selected.append(element_id)

    def _clear(self, attributes=True, metrics=True, attr_elements=True):
        """Removes all previously chosen objects from the filter."""
        if attributes:
            self.__clear_attr()
        if metrics:
            self.__clear_metr()
        if attr_elements:
            self.__clear_attr_elem()

    def __clear_attr(self):
        """Removes all previously chosen attributes from the filter."""
        self.attr_selected = []

    def __clear_metr(self):
        """Removes all previously chosen metrics from the filter."""
        self.metr_selected = []

    def __clear_attr_elem(self):
        """Removes all previously chosen attribute elements from the filter."""
        self.attr_elem_selected = []

    def _requested_objects(self):
        ro = {}

        ro = {"attributes": []}
        for i in self.attr_selected:
            if isinstance(i, list):
                forms = [{"id": form} for form in i[1:]]
                ro["attributes"].append({"id": i[0], "forms": forms})
            else:
                ro["attributes"].append({"id": i})

        ro["metrics"] = [{"id": i} for i in self.metr_selected]

        return ro

    def _view_filter(self):
        if not self.attr_elem_selected:
            return None

        else:
            # build {attribute_id:[element_id, element_id_n]} lookup dict
            lkp = {}
            for s in self.attr_elem_selected:
                if self.attr_elems[s]["attribute_id"] in lkp:
                    lkp[self.attr_elems[s]["attribute_id"]].append(s)
                else:
                    lkp[self.attr_elems.get(s)["attribute_id"]] = [s]

            # build view filter
            opers = []
            for k, v in lkp.items():
                att = {"type": "attribute", "id": k}
                elem = {"type": "elements", "elements": [{"id": _} for _ in v]}
                opers.append({"operator": self.operator, "operands": [att, elem]})

            if len(opers) > 1:
                vf = {"operator": "And", "operands": opers}
            else:
                vf = opers[0]

            return vf

    def _request_body(self):
        fb = {"requestedObjects": self._requested_objects()}
        if self.attr_elem_selected:
            fb["viewFilter"] = self._view_filter()

        return fb

    def _populate_attr_elements(self, attr_elements):
        for att in attr_elements:
            for el in att["elements"]:
                self.attr_elems[el["id"]] = {
                    "name": att["attribute_name"],
                    "attribute_id": att["attribute_id"],
                }

    def __type(self, object_id):
        """Look up and return object type from available objects."""

        if object_id in self.attributes:
            return "attribute"
        elif object_id in list(self.metrics) + self.row_count_metrics:
            return "metric"
        elif object_id in self.attr_elems:
            return "element"
        else:
            return None

    def __invalid(self, object_id):
        """Check if requested object_id is a valid object id."""
        object_is_attr_el = ':' in object_id
        if object_is_attr_el:
            return object_id.split(':')[0] not in self.attributes.keys()
        else:
            valid_object_ids = (
                list(self.metrics.keys())
                + list(self.attributes.keys())
                + self.row_count_metrics
            )
            return object_id not in valid_object_ids

    def __duplicated(self, object_id):
        """Check if requested object_id is already selected."""
        attr_selected = [elem[0] for elem in self.attr_selected]
        all_selected_objects = (
            attr_selected + self.metr_selected + self.attr_elem_selected
        )

        return object_id in all_selected_objects



from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error getting Drivers.")
def get_drivers(
    connection: Connection,
    error_msg: str | None = None,
):
    """Get information for all drivers.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """

    return connection.get(
        endpoint='/api/drivers',
    )


@ErrorHandler(err_msg="Error getting Driver with ID {id}")
def get_driver(connection: Connection, id: str, error_msg: str | None = None):
    """Get driver by a specific ID.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the driver
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    return connection.get(
        endpoint=f'/api/drivers/{id}',
    )


@ErrorHandler(err_msg="Error updating Driver with ID {id}")
def update_driver(
    connection: Connection,
    id: str,
    body: dict,
    error_msg: str | None = None,
):
    """Update a driver.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the driver
        body: Driver update info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    return connection.patch(
        endpoint=f'/api/drivers/{id}',
        json=body,
    )



import logging
from enum import auto
from typing import TYPE_CHECKING, Optional

from mstrio import config
from mstrio.api import datasources
from mstrio.datasources import DatasourceConnection, Dbms, list_datasource_connections
from mstrio.server.project import Project
from mstrio.users_and_groups.user import User
from mstrio.utils import helper
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, ObjectTypes
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.helper import (
    get_args_from_func,
    get_default_args_from_func,
    get_objects_id,
)
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.translation_mixin import TranslationMixin
from mstrio.utils.version_helper import class_version_handler, method_version_handler
from mstrio.utils.vldb_mixin import ModelVldbMixin

if TYPE_CHECKING:
    from mstrio.connection import Connection

logger = logging.getLogger(__name__)


@method_version_handler('11.3.0000')
def list_datasource_instances(
    connection: "Connection",
    to_dictionary: bool = False,
    limit: int | None = None,
    ids: list[str] | None = None,
    database_types: list[str] | None = None,
    project: Optional['Project | str'] = None,
    **filters,
) -> list["DatasourceInstance"] | list[dict]:
    """Get list of DatasourceInstance objects or dicts. Optionally filter the
    datasource instances by specifying filters.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        to_dictionary: If True returns dict, by default (False) returns
            User objects.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        ids: list of datasources ids as strings. By default `None`.
        database_types: list of strings representing database types.
            By default `None`. Available values: ['reserved', 'access',
            'altibase', 'amazon_athena', 'amazon_aurora', 'amazon_redshift',
            'arcadia_platform', 'aster', 'big_data_engine', 'cassandra',
            'cirro', 'cloud_element', 'cloud_gateway', 'cloud_gateway_aws_s3',
            'cloud_gateway_azure_adls_2', 'cloud_gateway_google_cloud_storage',
            'composite', 'concur', 'connection_cloud', 'data_direct_cloud',
            'datallegro', 'db2', 'denodo', 'drill', 'dropbox', 'eloqua',
            'enterprise_db', 'ess_base', 'exa_solution', 'excel', 'facebook',
            'gbase_8a', 'generic', 'generic_data_connector', 'google_analytics',
            'google_big_query', 'google_big_query_ff_sql', 'google_drive',
            'hive', 'hive_thrift', 'hubspot', 'impala', 'informatica',
            'informix', 'kognitiowx2', 'kyvos_mdx', 'mapd', 'marketo',
            'mark_logic', 'mem_sql', 'metamatrix', 'microsoft_as', 'mongo_db',
            'my_sql', 'neoview', 'netezza', 'open_access', 'oracle',
            'par_accel', 'par_stream', 'paypal', 'phoenix', 'pig',
            'pivotal_hawq', 'postgre_sql', 'presto', 'red_brick', 'salesforce',
            'sand', 'sap', 'sap_hana', 'sap_hana_mdx', 'search_engine',
            'servicenow', 'shopify', 'snow_flake', 'spark_config', 'spark_sql',
            'splunk', 'sql_server', 'square', 'sybase', 'sybase_iq',
            'sybase_sql_any', 'tandem', 'teradata', 'tm1', 'twitter', 'unknown',
            'url_auth', 'vectorwise', 'vertica', 'xquery']
        project: id (str) of a project or instance of an Project class
            to search for the datasource instances in. When provided, both
            `ids` and `database_types` are ignored. By default `None`.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'acg', 'datasource_type',
            'table_prefix', 'odbc_version', 'intermediate_store_db_name',
            'intermediate_store_table_space_name', 'dbms', 'owner', 'hidden',
            'datasource_connection', 'database_type', 'database_version',
            'primary_datasource', 'data_mart_datasource']

    Examples:
        >>> list_datasource_instances(connection, name='ds_instance_name')
    """
    return DatasourceInstance._list_datasource_instances(
        connection=connection,
        ids=ids,
        database_types=database_types,
        to_dictionary=to_dictionary,
        limit=limit,
        project=project,
        **filters,
    )


def list_connected_datasource_instances(
    connection: "Connection", to_dictionary: bool = False
) -> list["DatasourceInstance"] | list[dict]:
    """List all datasource instances for which there is an associated
    Database Login and are connected to a project mapped to a Connection object.

    Args:
        connection (Connection): MicroStrategy connection object returned by
            `connection.Connection()`
        to_dictionary (bool, optional): If True returns a list of dictionaries
            representing datasource instances

    Returns:
        list["DatasourceInstance"] | list[dict]: A list of connected
            datasource instances.
    """
    all_datasource_instances = list_datasource_instances(connection, to_dictionary=True)
    datasource_connections_ids = [
        ds_connection.get('id')
        for ds_connection in list_datasource_connections(connection, to_dictionary=True)
    ]
    connected_datasource_instances = [
        ds_instance
        for ds_instance in all_datasource_instances
        if ds_instance.get('datasource_connection').get('id')
        in datasource_connections_ids
        # remove xquery datasources because they are not available
        # in Workstation and listing namespaces for them can cause
        # IServer to become unresponsive.
        and ds_instance.get('database_type') != 'xquery'
    ]

    if to_dictionary:
        return connected_datasource_instances
    return [
        DatasourceInstance.from_dict(source=ds_instance, connection=connection)
        for ds_instance in connected_datasource_instances
    ]


class DatasourceType(AutoName):
    RESERVED = auto()
    NORMAL = auto()
    DATA_IMPORT = auto()
    DATA_IMPORT_PRIMARY = auto()


@class_version_handler('11.3.0000')
class DatasourceInstance(
    Entity, CopyMixin, DeleteMixin, ModelVldbMixin, TranslationMixin
):
    """Object representation of MicroStrategy DataSource Instance object.

    Attributes:
        connection: A MicroStrategy connection object
        id: Datasource Instance ID
        name: Datasource Instance name
        description: Datasource Instance description
        dbms: The database management system (DBMS) object
        database: `Database` object
        datasource_type: DatasourceType Enum (reserved, normal, data_import,
            data_import_primary)
        table_prefix: Table prefix
        odbc_version: Odbc version ENUM
        intermediate_store_db_name: Intermediate store DBName
        intermediate_store_table_space_name: intermediate store table space
            name
        type: Object type
        subtype: Object subtype
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        owner: User object that is the owner
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
    """

    _OBJECT_TYPE = ObjectTypes.DBROLE
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'dbms': Dbms.from_dict,
        'owner': User.from_dict,
        'datasource_type': DatasourceType,
        'datasource_connection': DatasourceConnection.from_dict,
    }
    _API_GETTERS = {
        (
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acl',
        ): objects_processors.get_info,
        (
            'id',
            'name',
            'description',
            'date_created',
            'date_modified',
            'datasource_type',
            'table_prefix',
            'odbc_version',
            'intermediate_store_db_name',
            'datasource_connection',
            'database_type',
            'database_version',
            'primary_datasource',
            'data_mart_datasource',
            'intermediate_store_table_space_name',
            'dbms',
            'acg',
        ): datasources.get_datasource_instance,
    }
    _API_PATCH: dict = {
        'abbreviation': (objects_processors.update, 'partial_put'),
        (
            "name",
            "description",
            "datasource_type",
            "table_prefix",
            "intermediate_store_db_name",
            "intermediate_store_table_space_name",
            "odbc_version",
            "datasource_connection",
            "primary_datasource",
            "data_mart_datasource",
            "dbms",
        ): (
            datasources.update_datasource_instance,
            'patch',
        ),
    }
    _PATCH_PATH_TYPES = {
        "name": str,
        "description": str,
        "datasource_type": str,
        "table_prefix": str,
        "intermediate_store_db_name": str,
        "intermediate_store_table_space_name": str,
        "odbc_version": str,
        "datasource_connection": dict,
        "primary_datasource": dict,
        "data_mart_datasource": dict,
        "dbms": dict,
    }
    _MODEL_VLDB_API = {
        'GET_ADVANCED': datasources.get_vldb_settings,
        'PUT_ADVANCED': datasources.update_vldb_settings,
        'GET_APPLICABLE': datasources.get_applicable_vldb_settings,
    }

    def __init__(
        self,
        connection: "Connection",
        name: str | None = None,
        id: str | None = None,
    ) -> None:
        """Initialize DatasourceInstance object by passing name or id.

        To explore all available DatasourceInstance objects use the
        `list_datasource_instance()` method.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            name: exact name of Datasource Instance
            id: ID of Datasource Instance
        """
        if id is None and name is None:
            helper.exception_handler(
                "Please specify either 'id' or 'name' parameter in the constructor."
            )

        if id is None:
            objects_info = DatasourceInstance._list_datasource_instances(
                connection=connection, name=name, to_dictionary=True
            )
            if objects_info:
                object_info, object_info["connection"] = objects_info[0], connection
                self._init_variables(**object_info)
            else:
                helper.exception_handler(
                    f"There is no Datasource Instance: '{name}'",
                    exception_type=ValueError,
                )
        else:
            super().__init__(connection=connection, object_id=id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.datasource_type = (
            DatasourceType(kwargs["datasource_type"])
            if kwargs.get("datasource_type")
            else None
        )
        self.table_prefix = kwargs.get("table_prefix")
        self.intermediate_store_db_name = kwargs.get("intermediate_store_db_name")
        self.intermediate_store_table_space_name = kwargs.get(
            "intermediate_store_table_space_name"
        )
        self.odbc_version = kwargs.get("odbc_version")
        self.datasource_connection = (
            DatasourceConnection.from_dict(
                kwargs.get("datasource_connection"), self.connection
            )
            if kwargs.get("datasource_connection")
            else None
        )
        self.database_type = kwargs.get("database_type")
        self.database_version = kwargs.get("database_version")
        self.primary_datasource = (
            DatasourceInstance.from_dict(
                kwargs.get("primary_datasource"), self.connection
            )
            if kwargs.get("primary_datasource")
            else None
        )
        self.data_mart_datasource = (
            DatasourceInstance.from_dict(
                kwargs.get("data_mart_datasource"), self.connection
            )
            if kwargs.get("data_mart_datasource")
            else None
        )
        self.dbms = (
            Dbms.from_dict(kwargs.get("dbms"), self.connection)
            if kwargs.get("dbms")
            else None
        )

    @classmethod
    def create(
        cls,
        connection: "Connection",
        name: str,
        dbms: Dbms | str,
        description: str | None = None,
        datasource_type: str | DatasourceType | None = None,
        table_prefix: str | None = None,
        odbc_version: str | None = None,
        intermediate_store_db_name: str | None = None,
        intermediate_store_table_space_name: str | None = None,
        datasource_connection: str | DatasourceConnection | None = None,
        database_type: str = None,
        database_version: str = None,
        primary_datasource: Optional["str | DatasourceInstance"] = None,
        data_mart_datasource: Optional["str | DatasourceInstance"] = None,
    ) -> Optional["DatasourceInstance"]:
        """Create a new DatasourceInstance object on I-Server.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            name: Datasource name
            dbms: The database management system (DBMS) object or id
            description: Datasource description
            datasource_type: DatasourceType Enum (reserved, normal, data_import,
                data_import_primary)
            table_prefix: Table prefix
            odbc_version: Odbc version ENUM (version3x, version2x)
            intermediate_store_db_name: Intermediate store DBName
            intermediate_store_table_space_name: intermediate store table space
                name
            datasource_connection: `DatasourceConnection` object or ID
            database_type: Database type
            database_version: Database version
            primary_datasource: `DatasourceInstance` object or ID
            data_mart_datasource: `DatasourceInstance` object or ID

        Returns:
            DatasourceInstance object.
        """
        dbms_id = get_objects_id(dbms, Dbms)
        connection_id = get_objects_id(datasource_connection, DatasourceConnection)
        primary_datasource_id = get_objects_id(primary_datasource, cls)
        data_mart_datasource_id = get_objects_id(data_mart_datasource, cls)
        database = {
            "type": database_type,
            "version": database_version,
            "connection": {"id": connection_id},
        }
        if primary_datasource_id:
            database["primaryDatasource"] = {"id": primary_datasource_id}
        if data_mart_datasource_id:
            database["dataMartDatasource"] = {"id": data_mart_datasource_id}

        body = {
            "name": name,
            "database": database,
            "description": description,
            "datasourceType": get_enum_val(datasource_type, DatasourceType),
            "tablePrefix": table_prefix,
            "odbcVersion": odbc_version,
            "intermediateStoreDbName": intermediate_store_db_name,
            "intermediateStoreTableSpaceName": intermediate_store_table_space_name,
            "dbms": {"id": dbms_id},
        }
        body = helper.delete_none_values(body, recursion=True)
        response = datasources.create_datasource_instance(connection, body).json()
        if config.verbose:
            logger.info(
                f"Successfully created datasource instance named: "
                f"'{response.get('name')}' with ID: '{response.get('id')}'"
            )
        return cls.from_dict(source=response, connection=connection)

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        datasource_type: str | DatasourceType | None = None,
        table_prefix: str | None = None,
        odbc_version: str | None = None,
        intermediate_store_db_name: str | None = None,
        intermediate_store_table_space_name: str | None = None,
        dbms: str | Dbms | None = None,
        datasource_connection: str | DatasourceConnection | None = None,
        primary_datasource: Optional["str | DatasourceInstance"] = None,
        data_mart_datasource: Optional["str | DatasourceInstance"] = None,
    ) -> None:
        """Alter DatasourceInstance properties.

        Args:
            name: Datasource name
            description: Datasource description
            datasource_type: DatasourceType Enum (reserved, normal, data_import,
                data_import_primary)
            table_prefix: Table prefix
            odbc_version: Odbc version ENUM (version3x, version2x)
            intermediate_store_db_name: Intermediate store DBName
            intermediate_store_table_space_name: intermediate store table space
                name
            dbms: The database management system (DBMS) object or ID
            datasource_connection: `DatasourceConnection` object or ID
            primary_datasource: `DatasourceInstance` object or ID
            data_mart_datasource: `DatasourceInstance` object or ID

        """
        dbms = {'id': get_objects_id(dbms, Dbms)} if dbms else None
        datasource_connection = (
            {'id': get_objects_id(datasource_connection, DatasourceConnection)}
            if datasource_connection
            else None
        )
        primary_datasource = (
            {'id': get_objects_id(primary_datasource, self)}
            if primary_datasource
            else None
        )
        data_mart_datasource = (
            {'id': get_objects_id(data_mart_datasource, self)}
            if data_mart_datasource
            else None
        )
        func = self.alter
        args = get_args_from_func(func)
        defaults = get_default_args_from_func(func)
        default_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    @classmethod
    def _list_datasource_instances(
        cls,
        connection: "Connection",
        to_dictionary: bool | None = False,
        limit: int | None = None,
        ids: list | None = None,
        database_types: list | None = None,
        project: Project | str | None = None,
        **filters,
    ) -> list["DatasourceInstance"] | list[dict]:
        project_id = project.id if isinstance(project, Project) else project
        objects = helper.fetch_objects(
            connection=connection,
            api=datasources.get_datasource_instances,
            dict_unpack_value="datasources",
            limit=limit,
            filters=filters,
            ids=ids,
            database_types=database_types,
            project=project_id,
        )
        if project_id:
            for obj in objects:
                obj['project_id'] = project_id

        if to_dictionary:
            return objects
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]

    def _set_object_attributes(self, **kwargs) -> None:
        super()._set_object_attributes(**kwargs)
        if kwargs.get("primary_datasource"):
            self.primary_datasource = DatasourceInstance(
                id=kwargs.get("primary_datasource").get("id"),
                name=kwargs.get("primary_datasource").get("name"),
                connection=self.connection,
            )
        if kwargs.get("data_mart_datasource"):
            self.data_mart_datasource = DatasourceInstance(
                id=kwargs.get("data_mart_datasource").get("id"),
                name=kwargs.get("data_mart_datasource").get("name"),
                connection=self.connection,
            )

    @method_version_handler('11.3.0900')
    def convert_ds_connection_to_dsn_less(self):
        """Convert datasource embedded connection from DSN to DSN-less format
        connection string and update the object to metadata.
        """
        response = datasources.convert_ds_dsn(
            connection=self.connection, datasource_id=self.id
        )
        if response.ok:
            self.fetch()



from typing import TYPE_CHECKING

from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.utils.sessions import FuturesSessionWithRenewal


@ErrorHandler(err_msg="Error getting user groups privileges for a group with ID {id}")
def get_privileges(
    connection, id, privilege_level=None, project_id=None, error_msg=None
):
    """Get user group's privileges of a project including the source of the
    privileges.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID of user group containing your required privileges
        privilege_level (string, optional) [server, project]: String
            representing level of privileges required
        project_id (string, optional): Project id string
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        endpoint=f'/api/usergroups/{id}/privileges',
        headers={'X-MSTR-ProjectID': None},
        params={'privilege.level': privilege_level, 'projectId': project_id},
    )


@ErrorHandler(err_msg="Error getting user groups memberships for a group with ID {id}")
def get_memberships(connection, id, error_msg=None):
    """Get information for the user group that is the direct parent of a
    specific user group.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID of user group containing your required memberships
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        endpoint=f'/api/usergroups/{id}/memberships',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error getting user groups members for a group with ID {id}")
def get_members(
    connection, id, include_access=False, offset=0, limit=-1, error_msg=None
):
    """Get member information for a specific user group.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID of user group containing your required privileges
        include_access (bool, optional): Specifies whether to return access for
            members
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        endpoint=f'/api/usergroups/{id}/members',
        headers={'X-MSTR-ProjectID': None},
        params={'includeAccess': include_access, 'offset': offset, 'limit': limit},
    )


@ErrorHandler(err_msg="Error getting user group settings for a group with ID {id}")
def get_settings(
    connection, id, include_access=False, offset=0, limit=-1, error_msg=None
):
    """Update the governing setting of the user group id.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID of user group containing your required privileges
        include_access (bool, optional): Specifies whether to return access for
            members
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/usergroups/{id}/settings',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error getting user groups top level.")
def get_top_level(connection, error_msg=None):
    """Get information for all of the user groups that exist at the level of
    the MicroStrategy Everyone user group.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        endpoint='/api/usergroups/topLevel',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error updating user group info with ID {id}")
def update_user_group_info(connection, id, body, error_msg=None):
    """Update specific information for a specific user group.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID of user group containing your required privileges
        body (JSON): Body.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object
    """

    return connection.patch(
        endpoint=f'/api/usergroups/{id}',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg="Error deleting user group with ID {id}")
def delete_user_group(connection, id, error_msg=None):
    """Delete user group for specific user group id.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID of user group containing your required privileges
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object
    """

    return connection.delete(
        endpoint=f'/api/usergroups/{id}',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error overwriting user group info for a group with ID {id}")
def replace_user_group_info(connection, id, error_msg=None):
    """Update all of the information for a specific user group.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID of user group containing your required privileges
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object
    """

    return connection.put(
        endpoint=f'/api/usergroups/{id}',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error getting information for a set of user groups.")
def get_info_all_user_groups(
    connection, name_begins, offset=0, limit=1000, fields=None, error_msg=None
):
    """Get information for a specific set of user groups or all user groups.
    You can refine the set of user groups that are returned with a query
    parameter that specifies the characters that the user group name must begin
    with; if you omit the nameBegins query parameter, all user groups are
    returned. You can limit the results that are returned with two query
    parameters. The offset parameter specifies the location in the list of
    users groups to begin returning results and the limit parameter specifies
    the maximum number of matching user groups that can be returned; if you set
    the limit parameter to -1, all results are returned (subject to Governing
    settings). The total count of results returned is included in the response
    header as "x-mstr-total-count". You obtain the authorization token needed
    to execute the request using POST /auth/login; you pass the authorization
    token in the request header.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        name_begins (string): only user groups with given beginning of name will
            be retrieved
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """

    return connection.get(
        endpoint='/api/usergroups/',
        headers={'X-MSTR-ProjectID': None},
        params={
            'nameBegins': name_begins,
            'offset': offset,
            'limit': limit,
            'fields': fields,
        },
    )


def get_info_all_user_groups_async(
    future_session: 'FuturesSessionWithRenewal',
    name_begins,
    offset=0,
    limit=-1,
    fields=None,
):
    """Get information for a set of users asynchronously.

    Args:
        future_session (FuturesSessionWithRenewal): Future Session object
            to call MicroStrategy REST Server asynchronously
        name_begins(string): Characters that the user name must begin with.
        offset(int): Starting point within the collection of returned search
            results. Used to control paging behavior.
        limit(int): Maximum number of items returned for a single search
            request. Used to control paging behavior. Use -1 (default ) for no
            limit (subject to governing settings).
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """

    params = {
        'nameBegins': name_begins,
        'offset': offset,
        'limit': limit,
        'fields': fields,
    }
    endpoint = '/api/usergroups/'
    headers = {'X-MSTR-ProjectID': None}
    future = future_session.get(endpoint=endpoint, headers=headers, params=params)
    return future


@ErrorHandler(err_msg="Error creating a new user group.")
def create_user_group(connection, body):
    """Create a new user group. The response includes the user group ID, which
    other endpoints use as a request parameter to specify the user group to
    perform an action on. You obtain the authorization token needed to execute
    the request using POST /auth/login; you pass the authorization token in the
    request header. You provide the information to create the.

    User group in the body parameter of the request - including the name and
    description of the user group, the IDs of parent user groups, and the IDs
    of child users or user groups.

    Args:
        connection: MicroStrategy REST API connection object
        body(JSON):{"name": "string",
                    "description": "string",
                    "memberships": ["string"],
                    "members": ["string"]}
    Returns:
        Complete HTTP response object
    """
    return connection.post(endpoint='/api/usergroups/', json=body)


@ErrorHandler(err_msg="Error getting user group information for a group with ID {id}")
def get_user_group_info(connection, id, error_msg=None):
    """Get information for a specific user group.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID of user group containing your required privileges
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        endpoint=f'/api/usergroups/{id}',
        headers={
            'X-MSTR-ProjectID': None,
        },
        params={
            'fields': 'id,name,type,subtype,ext_type,abbreviation,dateCreated,'
            'dateModified,version,owner,ancestors,acg,acl,ldapdn',
        },
        # we do not want to get information about members from
        # this endpoint as this is not enough to differentiate between users
        # and user groups
    )


@ErrorHandler(
    err_msg="Error getting user group security roles for a group with ID {id}"
)
def get_security_roles(connection, id, project_id=None, error_msg=None):
    """Get security roles for a specific user group in a specific project.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID of user group containing your required privileges
        project_id (string, optional): Project id string
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        endpoint=f'/api/usergroups/{id}/securityRoles',
        headers={'X-MSTR-ProjectID': None},
        params={'projectId': project_id},
    )


@ErrorHandler(err_msg="Error getting security filters for user with ID {id}.")
def get_security_filters(
    connection: 'Connection',
    id: str,
    projects: str | list[str] | None = None,
    offset: int = 0,
    limit: int = -1,
    error_msg: str | None = None,
):
    """Get each project level security filter and its corresponding inherited
    security filters for the user group with given ID.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): User group ID
        projects (str or list of str, optional): collection of projects' ids
            which is used for filtering data
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is -1.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    endpoint = f'/api/usergroups/{id}/securityFilters'

    if projects and isinstance(projects, list):
        projects = ','.join(projects)

    params = {'projects.id': projects, 'offset': offset, 'limit': limit}
    return connection.get(endpoint=endpoint, params=params)



import logging
from dataclasses import dataclass

from pandas import DataFrame, concat

from mstrio import config
from mstrio.api import documents
from mstrio.connection import Connection
from mstrio.object_management import Folder, SearchPattern, search_operations
from mstrio.project_objects.document import Document
from mstrio.server.environment import Environment
from mstrio.types import ObjectSubTypes
from mstrio.users_and_groups import UserOrGroup
from mstrio.utils import helper
from mstrio.utils.cache import CacheSource
from mstrio.utils.helper import Dictable, get_valid_project_id, is_dossier

logger = logging.getLogger(__name__)


helper.deprecation_warning(
    "mstrio.project_objects.dossier", "mstrio.project_objects.dashboard", "11.5.03"
)


def list_dossiers(
    connection: Connection,
    name: str | None = None,
    to_dictionary: bool = False,
    to_dataframe: bool = False,
    limit: int | None = None,
    project_id: str | None = None,
    project_name: str | None = None,
    **filters,
) -> list["Dossier"] | list[dict] | DataFrame:
    """Get all Dossiers stored on the server.

    Optionally use `to_dictionary` or `to_dataframe` to choose output format.
    If `to_dictionary` is True, `to_dataframe` is omitted.

    Args:
        connection(object): MicroStrategy connection object returned
            by 'connection.Connection()'
        name: characters that the dossier name must contain
        to_dictionary(bool, optional): if True, return Dossiers as
            list of dicts
        to_dataframe(bool, optional): if True, return Dossiers as
            pandas DataFrame
        limit(int): limit the number of elements returned. If `None` (default),
            all objects are returned.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        **filters: Available filter parameters: ['name', 'id', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'acg',
            'owner', 'ext_type', 'view_media', 'certified_info', 'project_id']

    Returns:
            List of dossiers or list of dictionaries or DataFrame object
    """

    return Dossier._list_all(
        connection,
        to_dictionary=to_dictionary,
        name=name,
        limit=limit,
        to_dataframe=to_dataframe,
        project_id=project_id,
        project_name=project_name,
        **filters,
    )


def list_dossiers_across_projects(
    connection: Connection,
    name: str | None = None,
    to_dictionary: bool = False,
    to_dataframe: bool = False,
    limit: int | None = None,
    **filters,
) -> list["Dossier"] | list[dict] | DataFrame:
    """Get all Dossiers stored on the server.

    Optionally use `to_dictionary` or `to_dataframe` to choose output format.
    If `to_dictionary` is True, `to_dataframe` is omitted.

    Args:
        connection(object): MicroStrategy connection object returned
            by 'connection.Connection()'
        name: characters that the dossier name must contain
        to_dictionary(bool, optional): if True, return Dossiers as
            list of dicts
        to_dataframe(bool, optional): if True, return Dossiers as
            pandas DataFrame
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['name', 'id', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'acg',
            'owner', 'ext_type', 'view_media', 'certified_info', 'project_id']

    Returns:
        List of dossiers or list of dictionaries or DataFrame object
    """
    project_id_before = connection.project_id
    env = Environment(connection)
    projects = env.list_projects()
    output = DataFrame() if to_dataframe else []
    for project in projects:
        try:
            connection.select_project(project_id=project.id)
        except ValueError:
            if config.verbose:
                logger.info(
                    f'Project {project.name} ({project.id}) is skipped '
                    f'because it does not exist or user has no access '
                    f'to it'
                )
            continue
        dossiers = Dossier._list_all(
            connection,
            to_dictionary=to_dictionary,
            name=name,
            limit=limit,
            to_dataframe=to_dataframe,
            **filters,
        )

        if to_dataframe:
            output = concat([output, dossiers], ignore_index=True)
        else:
            output.extend(dossiers)

    connection.select_project(project_id=project_id_before)
    return output[:limit]


class Dossier(Document):
    _CACHE_TYPE = CacheSource.Type.DOSSIER

    _API_GETTERS = {
        **Document._API_GETTERS,
        ('chapters', 'current_chapter'): documents.get_dossier_hierarchy,
    }
    _FROM_DICT_MAP = {
        **Document._FROM_DICT_MAP,
        'chapters': (
            lambda source, connection: [
                DossierChapter.from_dict(content, connection) for content in source
            ]
        ),
    }

    def __init__(
        self,
        connection: Connection,
        name: str | None = None,
        id: str | None = None,
    ):
        """Initialize Dossier object by passing name or id.

        Args:
            connection (object): MicroStrategy connection object returned
                by `connection.Connection()`
            name (string, optional): name of Dossier
            id (string, optional): ID of Dossier
        """
        super().__init__(connection=connection, name=name, id=id)

    def _init_variables(self, default_value, **kwargs):
        super()._init_variables(default_value=default_value, **kwargs)
        self._current_chapter = kwargs.get('current_chapter', default_value)
        self._chapters = kwargs.get('chapters', default_value)

    @classmethod
    def _list_all(
        cls,
        connection: Connection,
        search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
        name: str | None = None,
        to_dictionary: bool = False,
        to_dataframe: bool = False,
        limit: int | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        **filters,
    ) -> list["Dossier"] | list[dict] | DataFrame:
        if to_dictionary and to_dataframe:
            helper.exception_handler(
                "Please select either to_dictionary=True or to_dataframe=True, but not "
                "both.",
                ValueError,
            )
        project_id = get_valid_project_id(
            connection=connection,
            project_id=project_id,
            project_name=project_name,
            with_fallback=not project_name,
        )

        objects = search_operations.full_search(
            connection,
            object_types=ObjectSubTypes.REPORT_WRITING_DOCUMENT,
            project=project_id,
            name=name,
            pattern=search_pattern,
            **filters,
        )
        dossiers = [obj for obj in objects if is_dossier(obj['view_media'])]
        dossiers = dossiers[:limit]

        if to_dictionary:
            return dossiers
        elif to_dataframe:
            return DataFrame(dossiers)
        else:
            return [
                cls.from_dict(
                    source=dossier, connection=connection, with_missing_value=True
                )
                for dossier in dossiers
            ]

    def list_properties(self) -> dict:
        """List properties for the dossier."""
        properties = super().list_properties()
        additional_values = {
            'chapters': self.chapters,
            'current_chapter': self.current_chapter,
        }
        properties.update(additional_values)
        return properties

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        folder_id: Folder | str | None = None,
        hidden: bool | None = None,
    ):
        """Alter Dossier name, description and/or folder id.

        Args:
            name (string, optional): new name of the Dossier
            description (string, optional): new description of the Dossier
            folder_id (string | Folder, optional): A globally unique identifier
                used to distinguish between metadata objects within the same
                project. It is possible for two metadata objects in different
                projects to have the same Object Id.
            hidden (bool, optional): specifies whether the dossier is hidden
        """
        super().alter(name, description, folder_id, hidden)

    def publish(self, recipients: UserOrGroup | list[UserOrGroup] | None = None):
        """Publish the dossier for authenticated user. If `recipients`
        parameter is specified publishes the dossier for the given users.

        Args:
            recipients(UserOrGroup | list[UserOrGroup], optional): list of users
                or user groups to publish the dossier to (can be a list of IDs
                or a list of User and UserGroup elements)
        """
        super().publish(recipients)

    def unpublish(self, recipients: UserOrGroup | list[UserOrGroup] | None = None):
        """Unpublish the dossier for all users it was previously published to.
        If `recipients` parameter is specified unpublishes the dossier for the
        given users.

        Args:
            recipients(UserOrGroup | list[UserOrGroup], optional): list of users
                or user groups to publish the dossier to (can be a list of IDs
                or a list of User and UserGroup elements)
        """
        super().unpublish(recipients)

    def share_to(self, users: UserOrGroup | list[UserOrGroup]):
        """Shares the dossier to the listed users' libraries.

        Args:
            users(UserOrGroup | list[UserOrGroup]): list of users or user
                groups to publish the dossier to (can be a list of IDs or a
                list of User and UserGroup elements).
        """
        self.publish(users)

    @property
    def chapters(self) -> list["DossierChapter"]:
        return self._chapters

    @property
    def current_chapter(self) -> str:
        return self._current_chapter


@dataclass
class VisualizationSelector(Dictable):
    """Object that describes a Visualization Selector

    Attributes:
        visualization_key (string): key/id of the selector
        selector_type (string): type of the selector
        current_selection (dict): current selection of the selector
        targets (list[dict], optional): list of the selector's targets"""

    visualization_key: str
    selector_type: str
    current_selection: dict
    targets: list[dict] | None = None

    def list_properties(self, camel_case=True) -> dict:
        """Lists properties of visualization selector."""
        return self.to_dict(camel_case=camel_case)


@dataclass
class PageSelector(Dictable):
    """Object that describes a Page Selector

    Attributes:
        key (string): key/id of the selector
        selector_type (string): type of the selector
        current_selection (dict): current selection of the selector
        source (dict, optional): source of the selector
        multi_selection_allowed (bool): whether multi selection is allowed,
            defaults to False
        has_all_option (bool): whether the selector has all options enabled,
            defaults to False
        display_style (string, optional): style of the selector display
        available_object_items (list[dict], optional): list of objects available
            for the selector
        targets (list[dict], optional): list of targets of the selector
        name (string, optional): name of the selector
        summary (string, optional): summary of the selector"""

    key: str
    selector_type: str
    current_selection: dict
    source: dict | None = None
    multi_selection_allowed: bool = False
    has_all_option: bool = False
    display_style: str | None = None
    available_object_items: list[dict] | None = None
    targets: list[dict] | None = None
    name: str | None = None
    summary: str | None = None

    def list_properties(self, camel_case=True) -> dict:
        """Lists properties of page selector."""
        return self.to_dict(camel_case=camel_case)


@dataclass
class PageVisualization(Dictable):
    """Object that describes a Visualization on a Page

    Attributes:
        key (string): key/id of the visualization
        visualization_type (string, optional): type of the visualization
        name (string, optional): name of the visualization
        selector (VisualizationSelector, optional):
            selector for the visualization"""

    _FROM_DICT_MAP = {'selector': VisualizationSelector.from_dict}

    key: str
    visualization_type: str | None = None
    name: str | None = None
    selector: VisualizationSelector | None = None

    def list_properties(self, camel_case=True) -> dict:
        """Lists properties of page visualization."""
        return self.to_dict(camel_case=camel_case)


@dataclass
class ChapterPage(Dictable):
    """Object that describes a Chapter Page

    Attributes:
        key (string): key/id of the page
        visualizations (list[PageVisualization]): list of visualizations on the
            page
        name (string, optional): name of the page
        selectors (list[PageSelector], optional): list of selectors on the
            page"""

    _FROM_DICT_MAP = {
        'visualizations': [PageVisualization.from_dict],
        'selectors': [PageSelector.from_dict],
    }

    key: str
    visualizations: list[PageVisualization]
    name: str | None = None
    selectors: list[PageSelector] | None = None

    def list_properties(self, camel_case=True) -> dict:
        """Lists properties of dossier chapter page."""
        return self.to_dict(camel_case=camel_case)


@dataclass
class DossierChapter(Dictable):
    """Object that describes a Dossier Chapter

    Attributes:
        key (string): key/id of the chapter
        pages (list[ChapterPage]): list of the chapter pages
        name (string, optional): name of the chapter
        filters: (dict, optional): filters for the chapter"""

    _FROM_DICT_MAP = {'pages': [ChapterPage.from_dict]}

    key: str
    pages: list[ChapterPage]
    name: str | None = None
    filters: list[dict] | None = None

    def list_properties(self, camel_case=True) -> dict:
        """Lists properties of dossier chapter."""
        return self.to_dict(camel_case=camel_case)



import logging
from datetime import datetime, timezone
from enum import auto

from mstrio import config
from mstrio.api import schedules
from mstrio.connection import Connection
from mstrio.distribution_services.event import Event
from mstrio.distribution_services.schedule import ScheduleEnums, ScheduleTime
from mstrio.users_and_groups.user import User
from mstrio.utils import helper
from mstrio.utils.entity import DeleteMixin, Entity, ObjectTypes
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.time_helper import (
    DatetimeFormats,
    map_datetime_to_str,
    map_str_to_datetime,
)
from mstrio.utils.translation_mixin import TranslationMixin
from mstrio.utils.version_helper import method_version_handler

logger = logging.getLogger(__name__)


@method_version_handler('11.3.0000')
def list_schedules(
    connection: Connection, to_dictionary: bool = False, limit: int = None, **filters
) -> list["Schedule"] | list[dict]:
    """List schedule objects or schedule dictionaries. Optionally filter list.

    Args:
        connection(object): MicroStrategy connection object returned by
            'connection.Connection()'
        to_dictionary(bool, optional): if True, return Schedules as
            list of dicts
        limit(int, optional): maximum number of schedules returned.
        **filters: Available filter parameters: ['name', 'id', 'description',
            'schedule_type', 'schedule_next_delivery', 'start_date', 'time',
             'expired', 'acg']
    Returns:
        list["Schedule"] | list[dict]: [description]
    """

    objects = helper.fetch_objects(
        connection=connection,
        api=schedules.list_schedules,
        limit=limit,
        filters=filters,
        dict_unpack_value='schedules',
    )

    if to_dictionary:
        return objects
    else:
        return [
            Schedule.from_dict(source=obj, connection=connection) for obj in objects
        ]


class Schedule(Entity, DeleteMixin, TranslationMixin):
    """Class representation of MicroStrategy Schedule object.

    Attributes:
        connection: A MicroStrategy connection object
        name: Schedule name
        id: Schedule ID
        description: Schedule description
        schedule_type: Schedule type
        schedule_next_delivery: Schedule next delivery date
        time(ScheduleTime): Details of time-based schedule
        event(Event): Details of event-based schedule
    """

    class ScheduleType(AutoName):
        """Class representation of a type of a Microstrategy Schedule."""

        EVENT_BASED = auto()
        TIME_BASED = auto()
        NONE = None

    _OBJECT_TYPE = ObjectTypes.SCHEDULE_TRIGGER

    _API_GETTERS: dict = {
        (
            'id',
            'name',
            'description',
            'expired',
            'schedule_type',
            'schedule_next_delivery',
            'start_date',
            'stop_date',
            'time',
            'event',
        ): schedules.get_schedule,
        (
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'comments',
        ): objects_processors.get_info,
    }

    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': User.from_dict,
        'schedule_type': ScheduleType,
        'time': ScheduleTime.from_dict,
        'event': Event.from_dict,
        'schedule_next_delivery': DatetimeFormats.YMDHMS,
        'start_date': DatetimeFormats.DATE,
        'stop_date': DatetimeFormats.DATE,
    }
    _API_PATCH: dict = {
        (
            'abbreviation',
            'comments',
        ): (objects_processors.update, 'partial_put'),
        (
            'name',
            'description',
            'schedule_type',
            'start_date',
            'stop_date',
            'time',
            'event',
        ): (schedules.update_schedule, 'put'),
    }

    _PATCH_PATH_TYPES = {
        'name': str,
        'description': str,
        'abbreviation': str,
        'schedule_type': str,
        'start_date': str,
        'stop_date': str,
        'time': dict,
        'event': dict,
    }

    @method_version_handler('11.3.0000')
    def __init__(
        self, connection: Connection, id: str = None, name: str = None
    ) -> None:
        """Initialize the Schedule object, populates it with I-Server data.
        Specify either `id` or `name`. When `id` is provided (not `None`),
        `name` is omitted.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`.
            id: Schedule ID
            name: Schedule name
        Raises:
            AttributeError: when id is None and name is None.
            ValueError: when there is no schedule with a given name on the
            server.
        """

        if id is None and name is None:
            raise AttributeError(
                "Please specify either 'name' or 'id' parameter in the constructor."
            )
        if id is None:
            objects_info = list_schedules(
                connection=connection, name=name, to_dictionary=True
            )
            if objects_info:
                object_info, object_info["connection"] = objects_info[0], connection
                self._init_variables(**object_info)
            else:
                raise ValueError(f"There is no schedule with the given name: '{name}'")
        else:
            super().__init__(connection, id, name=name)

    @method_version_handler('11.3.0000')
    def _init_variables(self, **kwargs):
        """
        Add Schedule attributes to _AVAILABLE_ATTRIBUTES map.
        Set attributes inherited from Entity object.
        Set attributes related to schedule details.

        Args:
            **kwargs: dict with information about schedule attributes.
        Returns:
            None
        """
        super()._init_variables(**kwargs)
        self.schedule_type = self.ScheduleType(kwargs.get('schedule_type'))
        self.time = (
            ScheduleTime.from_dict(kwargs.get('time'))
            if (
                self.schedule_type == self.ScheduleType.TIME_BASED
                and kwargs.get('time')
            )
            else None
        )
        self.event = (
            Event.from_dict(kwargs.get('event'), connection=self._connection)
            if (
                self.schedule_type == self.ScheduleType.EVENT_BASED
                and kwargs.get('event')
            )
            else None
        )
        self._expired = kwargs.get('expired')
        self._schedule_next_delivery = map_str_to_datetime(
            "schedule_next_delivery",
            kwargs.get("schedule_next_delivery"),
            self._FROM_DICT_MAP,
        )
        self.start_date = map_str_to_datetime(
            "start_date", kwargs.get("start_date"), self._FROM_DICT_MAP
        )

    @method_version_handler('11.3.0000')
    def enable(self, stop_date: str | datetime) -> bool:
        """Enables schedule and sets stop date

        Args:
            stop_date: stop date provided either as a datetime or
            as a string in yyyy-MM-dd format
        Returns:
            Returns `True` if enabled properly, else `False`.
        """
        self._alter_properties(
            stop_date=map_str_to_datetime('stop_date', stop_date, self._FROM_DICT_MAP)
        )

        if config.verbose and not self.expired:
            logger.info(
                f'Schedule \'{self.name}\' with ID {self.id} has been enabled, '
                f'until {self.stop_date}.'
            )
            return True
        elif config.verbose:
            logger.info(
                f'Schedule \'{self.name}\' with ID {self.id} has NOT been enabled.'
            )
            return False

    @method_version_handler('11.3.0000')
    def disable(self, stop_date: str | datetime | None = None) -> bool:
        """Disable the schedule. Optional `stop_date` sets the date when
            the schedule should be disabled.

        Args:
            stop_date: stop date provided either as a datetime or
            as a string in yyyy-MM-dd format
        Returns:
            Returns `True` if disabled properly. It does not mean that schedule
            is already expired, as it can take up to one day.
            If operation failed, return `False`.
        """
        stop_date = (
            datetime.now(timezone.utc)
            if stop_date is None
            else map_str_to_datetime('stop_date', stop_date, self._FROM_DICT_MAP)
        )
        self._alter_properties(stop_date=stop_date)

        if config.verbose and self.expired:
            logger.info(
                f'Schedule \'{self.name}\' with ID {self.id} has been disabled.'
            )
            return True
        elif config.verbose and self.stop_date.date() == stop_date.date():
            logger.info(
                f"Schedule '{self.name}' with ID '{self.id}' has been set for "
                f"disabling. Depending on the schedule configuration (`event`, `time` "
                f"and `stop_date`), it will be disabled by day after "
                f"'{self.stop_date.date()}'."
            )
            return True
        else:
            logger.info(
                f"Schedule '{self.name}' with ID '{self.id}' has NOT been disabled."
            )
            return False

    @method_version_handler('11.3.0000')
    def list_properties(self):
        """List all properties of the object."""
        attributes = {
            key: self.__dict__[key] for key in self.__dict__ if not key.startswith('_')
        }
        attributes = {
            **attributes,
            'expired': self._expired,
            'schedule_next_delivery': self._schedule_next_delivery,
            'id': self.id,
            'type': self.type,
            'subtype': self.subtype,
            'ext_type': self.ext_type,
            'date_created': self.date_created,
            'date_modified': self.date_modified,
            'version': self.version,
            'owner': self.owner,
            'icon_path': self.icon_path,
            'view_media': self.view_media,
            'ancestors': self.ancestors,
            'certified_info': self.certified_info,
            'acg': self.acg,
            'acl': self.acl,
        }

        return {
            key: attributes[key]
            for key in sorted(attributes, key=helper.sort_object_properties)
        }

    @classmethod
    @method_version_handler('11.3.0000')
    def create(
        cls,
        connection: Connection,
        name: str,
        schedule_type: ScheduleType | str,
        start_date: str | datetime,
        description: str | None = None,
        stop_date: str | datetime | None = None,
        event_id: str | None = None,
        time: ScheduleTime | None = None,
        recurrence_pattern: ScheduleEnums.RecurrencePattern | str | None = None,
        execution_pattern: ScheduleEnums.ExecutionPattern | str | None = None,
        execution_time: str | None = None,
        start_time: str | None = None,
        stop_time: str | None = None,
        execution_repeat_interval: int | None = None,
        daily_pattern: ScheduleEnums.DailyPattern | str | None = None,
        repeat_interval: int | None = None,
        days_of_week: list[ScheduleEnums.DaysOfWeek] | list[str] | None = None,
        day: int | None = None,
        month: int | None = None,
        week_offset: ScheduleEnums.WeekOffset | str | None = None,
        day_of_week: ScheduleEnums.DaysOfWeek | str | None = None,
        weekday_off_set: str | None = None,
        days_of_month: list[str] | None = None,
        monthly_pattern: ScheduleEnums.MonthlyPattern | str | None = None,
        yearly_pattern: ScheduleEnums.YearlyPattern | str | None = None,
    ):
        """Create a Schedule using provided parameters as data.

        Args:
            recurrence_pattern (ScheduleEnums.RecurrencePattern, optional):
                The recurrence pattern of the schedule. Possible values are
                DAILY, WEEKLY, MONTHLY, YEARLY. Defaults to None.
            execution_pattern (ScheduleEnums.ExecutionPattern, optional):
                The execution pattern of the schedule. Possible values are ONCE,
                REPEAT. Defaults to None.
            execution_time (str, optional):
                The execution time of the execution day, if execution_pattern
                is ONCE. Format should be HH:mm:ss. Defaults to None.
            start_time (str, optional):
                The start time of the execution day, if execution_pattern is
                REPEAT. Format should be HH:mm:ss. Defaults to None.
            stop_time (str, optional):
                The stop time of the execution day, if execution_pattern is
                REPEAT.Format should be HH:mm:ss. Defaults to None.
            execution_repeat_interval (int, optional):
                The repeat interval of minutes of the execution day, if
                execution_pattern is REPEAT. Defaults to None.
            daily_pattern (ScheduleEnums.DailyPattern, optional):
                The daily recurrence pattern of the schedule. Possible values
                are DAY, WEEKDAY. Defaults to None.
            repeat_interval (int, optional):
                The repeat interval of days of daily schedule, if daily_pattern
                is DAY. Defaults to None.
            day (int, optional):
                The day in month of monthly schedule, if monthly_pattern is DAY
                or, The day in month of yearly schedule, if yearly_pattern is
                DAY. Defaults to None.
            month (int, optional):
                The month in year of yearly schedule. Defaults to None.
            week_offset (ScheduleEnums.WeekOffset, optional):
                The week offset in month of monthly schedule, if monthly_pattern
                is DAY_OF_WEEK or, The week offset in year of yearly schedule,
                if yearly_pattern is DAY_OF_WEEK. Possible values are FIRST,
                SECOND, THIRD, FOURTH, LAST. Defaults to None.
            day_of_week (ScheduleEnums.DaysOfWeek, optional):
                The days of week of weekly schedule or, The day of week in month
                of monthly schedule, if monthly_pattern is DAY_OF_WEEK or, The
                day of week in year of yearly schedule, if yearly_pattern is
                DAY_OF_WEEK. Possible values are: MONDAY, TUESDAY, WEDNESDAY,
                THURSDAY, FRIDAY, SATURDAY, SUNDAY. Defaults to None.
            weekday_offset (ScheduleEnums.WeekdayOffset, optional):
                The weekday offset in month of monthly schedule, if
                monthly_pattern is WEEKDAY. Defaults to None.
            days_of_month (List[str], optional):
                The days of month of monthly schedule, if monthly_pattern is
                DAYS_OF_MONTH. Must be provided as a list of one or more
                stringified digits (from '1' to '31'). Defaults to None.
            monthly_pattern (ScheduleEnums.MonthlyPattern, optional):
                The monthly recurrence pattern of the schedule. Possible values
                are: DAY, DAY_OF_WEEK, WEEKDAY, LAST_DAY, DAYS_OF_MONTH.
                Defaults to None.
            yearly_pattern (ScheduleEnums.YearlyPattern, optional):
                The yearly recurrence pattern of the schedule. Possible values
                are DAY, DAY_OF_WEEK. Defaults to None.
        Returns:
            Schedule object with provided parameters.
        """
        time_kwargs = {
            key: val
            for key, val in locals().items()
            if val is not None
            and key
            not in [
                'event_id',
                'connection',
                'description',
                'name',
                'schedule_type',
                'start_date',
                'stop_date',
                'time',
                'self',
                'cls',
            ]
        }
        # Event based or Time based logic
        if schedule_type == cls.ScheduleType.EVENT_BASED:
            execution_details = {'type': 'event', 'content': {'id': event_id}}
        elif schedule_type == cls.ScheduleType.TIME_BASED:
            if time is None:
                time = ScheduleTime.from_details(**time_kwargs)
            execution_details = {'type': 'time', 'content': time.to_dict()}

        # Datetime dates to string format conversion
        start_date = map_datetime_to_str(
            name='start_date', date=start_date, string_to_date_map=cls._FROM_DICT_MAP
        )
        stop_date = map_datetime_to_str(
            name='stop_date', date=stop_date, string_to_date_map=cls._FROM_DICT_MAP
        )

        # Create body and send request
        body = {
            'name': name,
            'description': description,
            'schedule_type': get_enum_val(schedule_type, cls.ScheduleType),
            'start_date': start_date,
            'stop_date': stop_date,
            execution_details['type']: execution_details['content'],
        }
        body = helper.delete_none_values(body, recursion=True)
        body = helper.snake_to_camel(body)
        # Response is already unpacked in wrapper
        response = schedules.create_schedule(connection, body)
        response = response.json()
        if config.verbose:
            logger.info(f"Created schedule '{name}' with ID: {response['id']}")
        return Schedule.from_dict(source=response, connection=connection)

    @method_version_handler('11.3.0000')
    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        start_date: str | datetime | None = None,
        stop_date: str | datetime | None = None,
        event: Event | None = None,
        event_id: str | None = None,
        time: ScheduleTime | None = None,
        recurrence_pattern: ScheduleEnums.RecurrencePattern | None = None,
        execution_pattern: ScheduleEnums.ExecutionPattern | None = None,
        execution_time: str | None = None,
        start_time: str | None = None,
        stop_time: str | None = None,
        execution_repeat_interval: int | None = None,
        daily_pattern: ScheduleEnums.DailyPattern | None = None,
        repeat_interval: int | None = None,
        days_of_week: list[ScheduleEnums.DaysOfWeek] | None = None,
        day: int | None = None,
        month: int | None = None,
        week_offset: ScheduleEnums.WeekOffset | None = None,
        day_of_week: ScheduleEnums.DaysOfWeek | None = None,
        weekday_offset: str | None = None,
        days_of_month: list[str] | None = None,
        monthly_pattern: ScheduleEnums.MonthlyPattern | None = None,
        yearly_pattern: ScheduleEnums.YearlyPattern | None = None,
        comments: str | None = None,
    ) -> None:
        """Alter Schedule properties.

        Args:
            recurrence_pattern (ScheduleEnums.RecurrencePattern, optional):
                The recurrence pattern of the schedule. Possible values are
                DAILY, WEEKLY, MONTHLY, YEARLY. Defaults to None.
            execution_pattern (ScheduleEnums.ExecutionPattern, optional):
                The execution pattern of the schedule. Possible values are ONCE,
                REPEAT. Defaults to None.
            execution_time (str, optional):
                The execution time of the execution day, if execution_pattern is
                ONCE. Format should be HH:mm:ss. Defaults to None.
            start_time (str, optional):
                The start time of the execution day, if execution_pattern is
                REPEAT. Format should be HH:mm:ss. Defaults to None.
            stop_time (str, optional):
                The stop time of the execution day, if execution_pattern is
                REPEAT. Format should be HH:mm:ss. Defaults to None.
            execution_repeat_interval (int, optional):
                The repeat interval of minutes of the execution day, if
                execution_pattern is REPEAT. Defaults to None.
            daily_pattern (ScheduleEnums.DailyPattern, optional):
                The daily recurrence pattern of the schedule. Possible values
                are DAY, WEEKDAY. Defaults to None.
            repeat_interval (int, optional):
                The repeat interval of days of daily schedule, if daily_pattern
                is DAY. Defaults to None.
            day (int, optional):
                The day in month of monthly schedule, if monthly_pattern is DAY
                or, The day in month of yearly schedule, if yearly_pattern is
                DAY. Defaults to None.
            month (int, optional):
                The month in year of yearly schedule. Defaults to None.
            week_offset (ScheduleEnums.WeekOffset, optional):
                The week offset in month of monthly schedule, if monthly_pattern
                is DAY_OF_WEEK or, The week offset in year of yearly schedule,
                if yearly_pattern is DAY_OF_WEEK. Possible values are FIRST,
                SECOND, THIRD, FOURTH, LAST. Defaults to None.
            day_of_week (ScheduleEnums.DaysOfWeek, optional):
                The days of week of weekly schedule or,
                The day of week in month of monthly schedule, if monthly_pattern
                is DAY_OF_WEEK or, The day of week in year of yearly schedule,
                if yearly_pattern is DAY_OF_WEEK. Possible values are: MONDAY,
                TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY, SUNDAY.
                Defaults to None.
            weekday_offset (ScheduleEnums.WeekdayOffset, optional):
                The weekday offset in month of monthly schedule, if
                monthly_pattern is WEEKDAY. Defaults to None.
            days_of_month (List[str], optional):
                The days of month of monthly schedule, if monthly_pattern is
                DAYS_OF_MONTH. Must be provided as a list of one or more
                stringified digits (from '1' to '31'). Defaults to None.
            monthly_pattern (ScheduleEnums.MonthlyPattern, optional):
                The monthly recurrence pattern of the schedule. Possible values
                are: DAY, DAY_OF_WEEK, WEEKDAY, LAST_DAY, DAYS_OF_MONTH.
                Defaults to None.
            yearly_pattern (ScheduleEnums.YearlyPattern, optional):
                The yearly recurrence pattern of the schedule. Possible values
                are DAY, DAY_OF_WEEK. Defaults to None.
            comments: long description of the schedule. Defaults to None.
        Returns:
            None
        """

        properties = {}

        # Event based or Time based logic
        if time is not None:
            properties['schedule_type'] = self.ScheduleType.TIME_BASED
            properties['time'] = time
        elif event is not None:
            properties['schedule_type'] = self.ScheduleType.EVENT_BASED
            properties['event'] = event
        elif event_id is not None:
            properties['schedule_type'] = self.ScheduleType.EVENT_BASED
            properties['event'] = Event(connection=self._connection, id=event_id)
        elif self.schedule_type == self.ScheduleType.TIME_BASED:
            properties['schedule_type'] = self.ScheduleType.TIME_BASED
            self.time.update_properties(
                recurrence_pattern,
                execution_pattern,
                execution_time,
                start_time,
                stop_time,
                execution_repeat_interval,
                daily_pattern,
                repeat_interval,
                days_of_week,
                day,
                month,
                week_offset,
                day_of_week,
                weekday_offset,
                days_of_month,
                monthly_pattern,
                yearly_pattern,
            )
            properties['time'] = self.time

        if name:
            properties['name'] = name
        if description:
            properties['description'] = description
        if start_date:
            properties['start_date'] = map_str_to_datetime(
                'start_date', start_date, self._FROM_DICT_MAP
            )
        if stop_date:
            properties['stop_date'] = map_str_to_datetime(
                'stop_date', stop_date, self._FROM_DICT_MAP
            )

        self._alter_properties(**properties)

        if (
            self.schedule_type == self.ScheduleType.EVENT_BASED
            and 'time' in self.__dir__()
        ):
            delattr(self, 'time')
        elif (
            self.schedule_type == self.ScheduleType.TIME_BASED
            and 'event' in self.__dir__()
        ):
            delattr(self, 'event')

    @method_version_handler('11.3.0000')
    def delete(self, force_with_dependents: bool = False) -> bool:
        """Delete the schedule.

        Args:
            force_with_dependents (bool, optional):
                If True, then no additional prompt will be shown before deleting
                the schedule. WARNING: This will also delete ALL SUBSCRIPTIONS
                that use this schedule. Defaults to False.
        Returns:
            bool: True if deletion was successful else False.
        """

        self._DELETE_CONFIRM_MSG = (
            "This schedule may be part of a subscription. "
            "Deleting such a schedule will remove the subscription as well. "
            "This action cannot be undone. "
            f"Are you sure you want to delete the schedule '{self.name}'"
            f" with ID: {self.id}?[Y/N]: "
        )
        self._DELETE_SUCCESS_MSG = f"Deleted schedule '{self.name}' with ID: {self.id}."

        return super().delete(force=force_with_dependents)

    @property
    def expired(self):
        """Whether or not the schedule is expired"""
        return self._expired

    @property
    def schedule_next_delivery(self):
        """Next delivery date of time based schedule"""
        return self._schedule_next_delivery



# flake8: noqa
from .attribute_form import AttributeForm
from .relationship import Relationship, RelationshipType

# isort: split

# This import has to be at the bottom due to circular import errors
from .attribute import Attribute, list_attributes



from enum import Enum


class PredefinedFolders(Enum):
    """Enumeration constants used to specify names of pre-defined folders.
    Values are specified according to `EnumDSSXMLFolderNames`."""

    AUTO_STYLES = 57
    BLACK_LISTED = 1000
    CONFIGURE_DB_ROLES = 73
    CONFIGURE_MONITORS = 58
    CONFIGURE_SERVER_DEFINITIONS = 59
    DB_CONNECTIONS = 81
    DB_LOGINS = 82
    DBMS = 76
    DEVICES = 88
    EVENTS = 72
    LINKS = 83
    LOCALES = 74
    MY_DOSSIERS = 96
    MY_SHARED_DOSSIERS = 97
    MY_DASHBOARDS = 96
    MY_SHARED_DASHBOARDS = 97
    PALETTES = 94
    PROFILE_ANSWERS = 21
    PROFILE_FAVORITES = 22
    PROFILE_OBJECTS = 19
    PROFILE_OTHER = 23
    PROFILE_REPORTS = 20
    PROFILE_SEGMENTS = 92
    PROJECTS = 77
    PROPERTY_SETS = 75
    PUBLIC_CONSOLIDATIONS = 2
    PUBLIC_CUSTOM_GROUPS = 3
    PUBLIC_FILTERS = 4
    PUBLIC_METRICS = 5
    PUBLIC_OBJECTS = 1
    PUBLIC_PROMPTS = 6
    PUBLIC_REPORTS = 7
    PUBLIC_SEARCHES = 8
    PUBLIC_TEMPLATES = 9
    ROOT = 39
    SCHEDULE_OBJECTS = 84
    SCHEDULE_TRIGGERS = 85
    SCHEMA_ARITHMETIC_OPERATORS = 49
    SCHEMA_ATTRIBUTE_FORMS = 25
    SCHEMA_ATTRIBUTES = 26
    SCHEMA_BASIC_FUNCTIONS = 41
    SCHEMA_COLUMNS = 27
    SCHEMA_COMPARISON_FOR_RANK_OPERATORS = 51
    SCHEMA_COMPARISON_OPERATORS = 50
    SCHEMA_DATA_EXPLORER = 28
    SCHEMA_DATE_AND_TIME_FUNCTIONS = 42
    SCHEMA_FACTS = 29
    SCHEMA_FINANCIAL_FUNCTIONS = 54
    SCHEMA_FUNCTIONS = 30
    SCHEMA_FUNCTIONS_NESTED = 40
    SCHEMA_HIERARCHIES = 31
    SCHEMA_INTERNAL_FUNCTIONS = 43
    SCHEMA_LOGICAL_OPERATORS = 52
    SCHEMA_MATH_FUNCTIONS = 55
    SCHEMA_NULL_ZERO_FUNCTIONS = 44
    SCHEMA_OBJECTS = 24
    SCHEMA_OLAP_FUNCTIONS = 45
    SCHEMA_OPERATORS = 48
    SCHEMA_PARTITION_FILTERS = 32
    SCHEMA_PARTITION_MAPPINGS = 33
    SCHEMA_PLUG_IN_PACKAGES = 53
    SCHEMA_RANK_AND_N_TILE_FUNCTIONS = 46
    SCHEMA_STATISTICAL_FUNCTIONS = 56
    SCHEMA_STRING_FUNCTIONS = 47
    SCHEMA_SUBTOTALS = 34
    SCHEMA_TABLES = 35
    SCHEMA_TRANSFORMATION_ATTRIBUTES = 37
    SCHEMA_TRANSFORMATIONS = 38
    SCHEMA_WAREHOUSE_TABLES = 36
    SECURITY_ROLES = 80
    SYSTEM_DIMENSION = 91
    SYSTEM_DRILL_MAP = 68
    SYSTEM_DUMMY_PARTITION_TABLES = 70
    SYSTEM_LINKS = 62
    SYSTEM_MD_SECURITY_FILTERS = 69
    SYSTEM_OBJECTS = 61
    SYSTEM_PARSER_FOLDER = 64
    SYSTEM_PROPERTY_SETS = 63
    SYSTEM_SCHEMA_FOLDER = 65
    SYSTEM_SYSTEM_HIERARCHY = 67
    SYSTEM_SYSTEM_PROMPTS = 71
    SYSTEM_WAREHOUSE_CATALOG = 66
    TABLE_SOURCES = 86
    TEMPLATE_ANALYSIS = 93
    TEMPLATE_CONSOLIDATIONS = 11
    TEMPLATE_CUSTOM_GROUPS = 12
    TEMPLATE_DASHBOARDS = 90
    TEMPLATE_DOCUMENTS = 60
    TEMPLATE_FILTERS = 13
    TEMPLATE_METRICS = 14
    TEMPLATE_OBJECTS = 10
    TEMPLATE_PROMPTS = 15
    TEMPLATE_REPORTS = 16
    TEMPLATE_SEARCHES = 17
    TEMPLATE_TEMPLATES = 18
    THEMES = 95
    TRANSMITTERS = 89
    USER_GROUPS = 79
    USERS = 78
    VERSION_UPDATE_HISTORY = 87



# This example is for a dashboard user to modify data in a PostgreSQL DB table using 
# a modern grid configured with transaction functions. The workflow will rely on this 
# script to make actual modifications in the table. The table named USERS has 4 
# columns USER_ID(Int), USER_LOGIN(Varchar), HIRED_DATE(Date) and EMAIL(VarChar) 
# with USER_ID as the primary. Data changes for HIRED_DATE (Date) and EMAIL(VarChar) 
# are optional.

# In this template we use the library 'psycopg2' (installed in runtime as
# 'psycopg2-binary') to connect with the PostgreSQL database, submit atomic requests 
# and execute operations. For using other types of database, make sure to use
# their methods correctly. 

# Python libraries to handle SQL connections are not installed on the environment by
# default so it is necessary to choose runtime which have such library installed in the
# settings of the script. Additionally it is necessary to specify correct `Network
# Access` to allow script to communicate with some other IP then IP of the environment.
import psycopg2
import datetime


# helper method to handle not adding None values to the query
def append_not_none_value(array, value):
    if value is not None:
        array.append(value)
        return True
    return False

# Helper method to return datetime object in correct format for SQL operations
def parseDateTimeValue(dateValue):
    # somehow it will be converted to this datetime format
    #   "%Y-%m-%d %H:%M:%S.%f" -> 2023-04-01 00:00:00.000, 
    date_formats = [
        "%Y-%m-%d %H:%M:%S.%f",
        "%m/%d/%Y", 
        "%m/%d/%Y %H:%M:%S",
        "%Y-%d-%m %H:%M:%S",
        "%m/%d/%Y %H:%M:%S.%f",
        "%m-%d-%Y"
    ]
    parsed_datetime = None
    for date_format in date_formats:
        try:
            parsed_datetime = datetime.datetime.strptime(dateValue, date_format)
            break
        except:
            pass

    if (parsed_datetime == None):
        raise Exception("Wrong Type of the datatime input value convertion!!!!")
    else:
        return parsed_datetime


try:
    # create connection to database. If you want to use other way then providing
    # user and password please check documentation of this method 
    connection = None
    connection = psycopg2.connect(
        host=$host,
        port=$port,
        database=$database,
        user=$user_name,
        password=$user_password
    )

    # Prepare query for database operations. Due to the fact that for some rows
    # we might want to pass `None` value we need some additional logic. This logic
    # require to create query during the time of reading passed properties.

    # Parameterized query for insert new row
    # We assume that we must pass `USER_ID` and `USER_LOGIN` when inserting
    # and the other properties are optional.
    tmp_sql_insert_query = """
        INSERT INTO USERS (
            USER_ID,
            USER_LOGIN
    """

    # create cursor to execute operations on database
    cursor = connection.cursor()
    
    # deal with the every row data via prepared statements
    # we assume that the logic of passing data from dashboard works correctly and each
    # transactional column is either the array of length which is equal to number
    # of rows passed from dashboard or `None`.
    for index in range(len($user_id)):
        # when value is optional and not provided, then we don't add it to insert query
        optional_values_array = []
        optional_values_in_insert_query = ''
        # this value depends on number of the properties in initial INSERT query
        # it is equal to the number of properties there
        i = 2
        # when array with values for each property is not `None` and the value
        # for particular row is not `None` then we add it
        if $hired_date and $hired_date[index]:
            optional_values_array.append(parseDateTimeValue($hired_date[index]))
            optional_values_in_insert_query += f",HIRED_DATE"
            i = i + 1
        if $email and append_not_none_value(optional_values_array, $email[index]):
            optional_values_in_insert_query += f",EMAIL"
            i = i + 1

        # create list of parameters
        query_parameters = ",".join(list(map(lambda x: "%s", list(range(i)))))
        optional_values_in_insert_query += f") VALUES ({query_parameters})"

        sql_insert_tuple = (
            int($user_id[index]),
            $user_login[index],
        ) + tuple(optional_values_array)
        # final INSERT query is created based on initial INSERT query and the part
        # created for particular row based on properties passed for this row
        sql_insert_query = tmp_sql_insert_query + optional_values_in_insert_query

        # execute insert query
        cursor.execute(sql_insert_query, sql_insert_tuple)

    # commit all changes when there weren't any errors
    connection.commit()

# error handling
except psycopg2.Error as e:
    # in case of any error we have to rollback all of the changes which
    # has been made on database
    if connection:
        connection.rollback()
    print("Error:", e)

finally:
    # close the database connection
    if connection:
        connection.close()
        


from mstrio.connection import get_connection
from mstrio.server import Project

conn = get_connection(workstationData, 'MicroStrategy Tutorial')

project = Project(connection=conn, id='B3FEE61A11E696C8BD0F0080EFC58F44')
test_project = Project(connection=conn, id='B7CA92F04B9FAE8D941C3E9B7E0CD754')

# Change single setting for a project
# You can get a list of all available settings by calling:
# project.settings.list_properties()
project.settings.maxMstrFileSize = 50
project.settings.update()

# Apply all settings from one project to another
project.settings.alter(**test_project.settings.list_properties(show_names=False))
project.settings.update()

# Or export settings to a CSV file and import them to another project
settings_file_name = 'settings.csv'
test_project.settings.to_csv(settings_file_name)
project.settings.import_from(settings_file_name)
project.settings.update()

# Export settings to a CSV file with their descriptions
project.settings.to_csv('settings_file_name', show_description=True)
